
// ----------------------------------------------------------------------

void appendOffset() {
  // this is called when Dither is active to append the offset to the
  // current servo tyarget
  if (comFlag < 1) {return;}  // don't do anything if not connected

  int zSV;
  setSliderX();

  zSV = servoVal[servoTgt] + DitherOffset;
  zSV = max(zSV,servoVLL[servoTgt]); zSV = min(zSV,servoVUL[servoTgt]);
  msgTx = "ST" + str(servoTgt) + ".SV" + str(zSV) + ".";
  usbPortWrite(msgTx); drawFlag = 1;
}

// ----------------------------------------------------------------------

void centreJoyX() {
  // centre the joystick based on the current limits, LL, UL
  servoVal[servoTgt] = servoVLL[servoTgt] + ((servoVUL[servoTgt] - servoVLL[servoTgt])/2);
  setSliderX(); drawFlag = 1;
}

// ----------------------------------------------------------------------

void centreServo() {
  // centre the servo based on the current limits, LL, UL
  centreJoyX(); servoLast = servoVal[servoTgt];
  msgTx = "SV" + str(servoVal[servoTgt]) + ".";
  usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

void ClearChEn() {
  // clear the channel enable flags
  for (int zC = 0; zC < 16; zC++) {Ch_En[zC] = false;}
  Ch_Cnt = 0;
}

// ----------------------------------------------------------------------

void ClearMemLamps() {
  // clear the memory state flags
  for (int zC = 0; zC < 16; zC++) {memLamp[zC] = 1;}
}

// ----------------------------------------------------------------------

void closeCOMPort() {
  // Close the COM port and empty the serial buffer
  if (comFlag > 0) {
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; drawFlag = 1;
    msgTx = "COM Port Closed"; println(msgTx);
    servoOE = 1;  // turn OFF OE enabled lamp
    ClearChEn();  // clear all of the channel ON flags
    Dither = false;
  }
}

// ----------------------------------------------------------------------

void copyToClipboard() {
  // called when user clicks on 'Copy' button
  // the code either copies a declaration of arrays or a list to the clipboard
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't copy if not in default mode

  int zS; String zH = ""; String zN = "";
  if (mB == LEFT) {
    // left mouse click so copy a declaration of arrays
    CopyData = "// Define PWM freq and servo calibration arrays" + CR;
    CopyData = CopyData + "#define Freq " + str(Freq) + "  // MG996R servos run at ~50 Hz updates" + CR;
    CopyData = CopyData + "// Servo:   { 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F }" + CR;
    CopyData = CopyData + "int SLL[] = {";
    for (zS = 0; zS < 16; zS++) {
      CopyData = CopyData + str(memVLL[zS]);
      if (zS < 15) {CopyData = CopyData + ",";}
    }
    CopyData = CopyData + "};  // servo lower limit values" + CR;
    CopyData = CopyData + "int SRV[] = {";
    for (zS = 0; zS < 16; zS++) {
      CopyData = CopyData + str(memVal[zS]);
      if (zS < 15) {CopyData = CopyData + ",";}
    }
    CopyData = CopyData + "};  // reference values used to set a servo" + CR;
    CopyData = CopyData + "int SUL[] = {";
    for (zS = 0; zS < 16; zS++) {
      CopyData = CopyData + str(memVUL[zS]);
      if (zS < 15) {CopyData = CopyData + ",";}
    }
    CopyData = CopyData + "};  // servo upper limit values" + CR;
    msgTx = "Arrays Copied"; drawTxMsg();
  } else {
    // right mouse click so copy a definition list
    CopyData = CR + "// Define servo Cal. constants using Hex references" + CR;
    CopyData = CopyData + "#define Freq " + str(Freq) + "  // MG996R servos run at ~50 Hz updates" + CR;
    // create lower limit list
    for (zS = 0; zS < 16; zS++) {
      zH = str(zS);
      if (zS == 10) {zH = "A";}
      if (zS == 11) {zH = "B";}
      if (zS == 12) {zH = "C";}
      if (zS == 13) {zH = "D";}
      if (zS == 14) {zH = "E";}
      if (zS == 15) {zH = "F";}
      zN = str(zS); if (zS < 10) {zN = "0" + zN;} 
      CopyData = CopyData + "#define SLL" + zH + " " + str(memVLL[zS]);
      CopyData = CopyData + "\t// servo " + zN + " lower limit value" + CR;
    }
    // create centre value list
    for (zS = 0; zS < 16; zS++) {
      zH = str(zS);
      if (zS == 10) {zH = "A";}
      if (zS == 11) {zH = "B";}
      if (zS == 12) {zH = "C";}
      if (zS == 13) {zH = "D";}
      if (zS == 14) {zH = "E";}
      if (zS == 15) {zH = "F";}
      zN = str(zS); if (zS < 10) {zN = "0" + zN;} 
      CopyData = CopyData + "#define SRV" + zH + " " + str(memVal[zS]);
      CopyData = CopyData + "\t// servo " + zN + " reset value" + CR;
    }
    // create upper limit list
    for (zS = 0; zS < 16; zS++) {
      zH = str(zS);
      if (zS == 10) {zH = "A";}
      if (zS == 11) {zH = "B";}
      if (zS == 12) {zH = "C";}
      if (zS == 13) {zH = "D";}
      if (zS == 14) {zH = "E";}
      if (zS == 15) {zH = "F";}
      zN = str(zS); if (zS < 10) {zN = "0" + zN;} 
      CopyData = CopyData + "#define SUL" + zH + " " + str(memVUL[zS]);
      CopyData = CopyData + "\t// servo " + zN + " upper limit value" + CR;
    }
    msgTx = "List Copied"; drawTxMsg();
  }
  // now copy it
  StringSelection data = new StringSelection(CopyData);
  Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
  clipboard.setContents(data, data);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void countChannels() {
  // called to get a handle on the number of active channels
  Ch_Cnt = 0;
  for (int zC = 0; zC < 16; zC++) {
    if (Ch_En[zC] == true) {Ch_Cnt++;}
  }
}

// ----------------------------------------------------------------------

void decActiveSliders() {
  // normally called from a keyboard left arrow key event.
  // It will decrement any active slider value
  int zservoTgt = servoTgt;  // note current target value
  for (servoTgt = 0; servoTgt < 16; servoTgt++) {
    if (Ch_En[servoTgt]) {decXSlider();}
  }
  servoTgt = zservoTgt;  // restore target value
}

// ----------------------------------------------------------------------

void decDither() {
  // reduce the dither value and send it to the Arduino
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't apply dither if not in default mode
  
  if (DitherVal > 1) {
    DitherVal--; drawFlag = 1;
    //msgTx = "SF" + str(Freq) + ".";
    //usbPortWrite(msgTx);
  }
  if (Dither) {DitherCnt = 50;  DitherPhase = 0;}  // display the change
}

// ----------------------------------------------------------------------

void decFreq() {
  // reduce the PWM frequency value and send it to the Arduino
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't dec freq. if not in default mode

  if (Freq > 40) {
    Freq--; drawFlag = 1;
    msgTx = "SF" + str(Freq) + ".";
    usbPortWrite(msgTx);
  }
}

// ----------------------------------------------------------------------

void decXSlider() {
  // decrement target slider value
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (servoVal[servoTgt] > servoVLL[servoTgt]) {
    servoLast = servoVal[servoTgt];
    if (keySHIFT) {servoVal[servoTgt] -= 5;} else {servoVal[servoTgt]--;}
    if (servoVal[servoTgt] < servoVLL[servoTgt]) {servoVal[servoTgt] = servoVLL[servoTgt];}
    setXSlider();
    if (AppMode == 0) {
      if (Ch_En[servoTgt]) {
        msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
      } else {
        msgTx = "ST" + str(servoTgt) + "." +"SV0.";
      }
      usbPortWrite(msgTx);
    }
    if (AppMode == 1) {
      msgTx = "ST" + str(servoTgt) + "." +"SA" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 2) {
      msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 3) {
      msgTx = "LT" + str(servoTgt) + "." +"LV" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    valueChanged();
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void drawBlankouts() {
  // in angle mode so blank out certain form functions
  stroke(255,255,255); strokeWeight(1);
  fill(255,255,255);    // set fill colour
  rect(276,7,423,38);   // hide top controls
  rect(781,31,32,636);  // hide side lamps
  rect(700,658,78,36);  // Copy button
  
  TR=0; TG=0; TB=0;
  if (AppMode == 1) {drawTxtField("- Angle Mode",278,17);}
  if (AppMode == 2) {drawTxtField("- PWM Mode",278,17);}
  if (AppMode == 3) {
    drawTxtField("- LX-16A Servo Mode",278,17);
    TB=255;  // revert to blue text
    // draw timer field
    stroke(0,0,0); strokeWeight(1);
    fill(255,255,255);    // set fill colour white
    rect(696,665,83,26);   // draw field window
    if (msRun) {
      // msTimer is running so display current time difference
      drawTxtCtrField(str(millis() - msTimer),737, 670);
    } else {
      // msTimer n ot running so display timer value
      drawTxtCtrField(str(msTimer),737, 670);
    }
  } TB=255;  // revert to blue text
}

// ----------------------------------------------------------------------

void drawCOMMsg() {
  // put text in COM value field
  if (comFlag < 1) {TR = 255;}  // show text in red if disconnected
  drawTxtCtrField(msgCOM, 147, 672); TR = 0;
}

// ----------------------------------------------------------------------

void drawDither() {
  // put text in the dither field
  if (AppMode > 0) {return;}  // don't draw if not in default mode

  if (comFlag < 1) {TR = 255; TG = 0; TB = 0;}
  else {
    if (Dither){TR = 0; TG = 0; TB = 255; BR = 255; BG = 255; BB = 0;}
    else {TR = 150; TG = 150; TB = 150; BR = 255; BG = 255; BB = 255;}
  }
  if ((comFlag > 0) && Dither) {
    // dither is active so display the changing offset value
    String zTx$ = "  ";
    if (DitherCnt > 25) {
      zTx$ += str(DitherVal);
    } else {
      if (DitherOffset > 0) {zTx$ = " +";}
      zTx$ += str(DitherOffset);
    }
    zTx$ += " "; drawTxtCtrField(zTx$, 588, 18);
  } else {
    // either no COM or Dither active so display value
    drawTxtCtrField("  " + str(DitherVal) +  "  ", 588, 18);
  } TR = 0; TG = 0; TB = 255; BR = 255; BG = 255; BB = 255;
}

// ----------------------------------------------------------------------

void drawFreqMsg() {
  // put text in the Frequency field
  if (AppMode > 0) {return;}  // don't draw if not in default mode

  if (comFlag < 1) {TR = 255; TG = 0; TB = 0;}
  drawTxtCtrField(str(Freq), 490, 18); TR = 0;
}

// ----------------------------------------------------------------------

void drawHelp() {
  // display help messages
  TR = 0; TG = 0; TB = 255;
  drawTxtCtrField(Help$, 408, 702);
}

// ----------------------------------------------------------------------

void drawJoyX() {
  // draw the X joystick slider
  if (comFlag < 1) {return;}  // don't draw if not connected
  if (!data_loaded && (AppMode == 0)) {return;}  // don't draw if no data loaded
  
  int zXw = 16; stroke(0,0,0); 
  strokeWeight(2);
  if (AppMode < 3) {  
    if (Ch_En[0]) {fill(255,100,100);} else {fill(255,255,255);}
    rect(JoyXval[0]-zXw,49,zXw+zXw,30);
  }
  
  if (Ch_En[1]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[1]-zXw,87,zXw+zXw,30);
  if (Ch_En[2]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[2]-zXw,125,zXw+zXw,31);
  if (Ch_En[3]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[3]-zXw,164,zXw+zXw,31);
  
  if (Ch_En[4]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[4]-zXw,202,zXw+zXw,31);
  if (Ch_En[5]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[5]-zXw,240,zXw+zXw,31);
  if (Ch_En[6]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[6]-zXw,279,zXw+zXw,31);
  if (Ch_En[7]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[7]-zXw,317,zXw+zXw,31);
  
  if (Ch_En[8]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[8]-zXw,356,zXw+zXw,31);
  if (Ch_En[9]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[9]-zXw,394,zXw+zXw,31);
  if (Ch_En[10]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[10]-zXw,432,zXw+zXw,31);
  if (Ch_En[11]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[11]-zXw,471,zXw+zXw,31);
  
  if (Ch_En[12]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[12]-zXw,509,zXw+zXw,31);
  if (Ch_En[13]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[13]-zXw,547,zXw+zXw,31);
  if (Ch_En[14]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[14]-zXw,586,zXw+zXw,31);
  if (Ch_En[15]) {fill(255,100,100);} else {fill(255,255,255);}
  rect(JoyXval[15]-zXw,625,zXw+zXw,31);
}

// ----------------------------------------------------------------------

void drawLLMsg() {
  // put text in LL value field
  if (comFlag < 1) {return;}  // don't draw if not connected
  int zX = 182;
  if (AppMode < 3) {drawTxtCtrField("       ", zX, 57); drawTxtCtrField(str(servoVLL[0]), zX, 57);}    // Ch0
  drawTxtCtrField("       ", zX, 95); drawTxtCtrField(str(servoVLL[1]), zX, 95);    // Ch1
  drawTxtCtrField("       ", zX, 134); drawTxtCtrField(str(servoVLL[2]), zX, 134);  // Ch2
  drawTxtCtrField("       ", zX, 172); drawTxtCtrField(str(servoVLL[3]), zX, 172);  // Ch3

  drawTxtCtrField("       ", zX, 211); drawTxtCtrField(str(servoVLL[4]), zX, 211);  // Ch4
  drawTxtCtrField("       ", zX, 249); drawTxtCtrField(str(servoVLL[5]), zX, 249);  // Ch5
  drawTxtCtrField("       ", zX, 288); drawTxtCtrField(str(servoVLL[6]), zX, 288);  // Ch6
  drawTxtCtrField("       ", zX, 326); drawTxtCtrField(str(servoVLL[7]), zX, 326);  // Ch7

  drawTxtCtrField("       ", zX, 364); drawTxtCtrField(str(servoVLL[8]), zX, 364);  // Ch8
  drawTxtCtrField("       ", zX, 402); drawTxtCtrField(str(servoVLL[9]), zX, 402);  // Ch9
  drawTxtCtrField("       ", zX, 440); drawTxtCtrField(str(servoVLL[10]), zX, 440); // Ch10
  drawTxtCtrField("       ", zX, 479); drawTxtCtrField(str(servoVLL[11]), zX, 479); // Ch11

  drawTxtCtrField("       ", zX, 518); drawTxtCtrField(str(servoVLL[12]), zX, 518);  // Ch12
  drawTxtCtrField("       ", zX, 556); drawTxtCtrField(str(servoVLL[13]), zX, 556);  // Ch13
  drawTxtCtrField("       ", zX, 594); drawTxtCtrField(str(servoVLL[14]), zX, 594);  // Ch14
  drawTxtCtrField("       ", zX, 632); drawTxtCtrField(str(servoVLL[15]), zX, 632);  // Ch15
}

// ----------------------------------------------------------------------

void drawMemlamps() {
  // draw Memory status lamps
  if (comFlag < 1) {return;}   // don't draw if not connected
  if (AppMode > 0) {return;}       // don't draw if in angle mode
  
  stroke(255,255,255); strokeWeight(1);
  for (int zS = 0; zS < 16; zS++) {
    switch(memLamp[zS]) {
      case 0: fill(128,128,128); break; // grey
      case 1: fill(40,210,0); break;    // green - reset value
      case 2: fill(220,220,40); break;  // yellow - modified from reset
      case 3: fill(64,64,255); break;   // blue - stored value
      case 4: fill(210,64,210); break;   // purple - modified from stored value
    }
    switch(zS) {
      case 0: rect(787,53,20,20); break;
      case 1: rect(787,92,20,20); break;
      case 2: rect(787,130,20,20); break;
      case 3: rect(787,168,20,20); break;
      case 4: rect(787,207,20,20); break;
      case 5: rect(787,246,20,20); break;
      case 6: rect(787,284,20,20); break;
      case 7: rect(787,322,20,20); break;
      case 8: rect(787,360,20,20); break;
      case 9: rect(787,399,20,20); break;
      case 10: rect(787,437,20,20); break;
      case 11: rect(787,475,20,20); break;
      case 12: rect(787,514,20,20); break;
      case 13: rect(787,552,20,20); break;
      case 14: rect(787,591,20,20); break;
      case 15: rect(787,629,20,20); break;
    }
  }
}

// ----------------------------------------------------------------------

void drawMouseXY() {
  // draws the value of mouse x,y top left
  // draws a full X/Y axis on the form
  if ((mX != mXL) || (mY != mYL)) {
    // draw full X/Y axis
    stroke(255,0,0,128); strokeWeight(1);
    line(0,mY,823,mY); line(mX,0,mX,731);

    int zX,zY;
    if (mX > 412) {zX = mX - 58;} else {zX = mX + 8;}
    if (mY > 380) {zY = mY - 38;} else {zY = mY + 8;}
    stroke(0,0,0); fill(255,255,255); rect(zX,zY,50,30);
    fill(0,0,0); text("X=" + mX, zX + 4, zY + 13); text("Y=" + mY, zX + 4, zY + 27);
    mXL = mX; mYL = mY;
  }
}

// ----------------------------------------------------------------------

void drawOElamp() {
  // draw OE lamp if active
  if (comFlag < 1) {return;}  // don't draw if not connected
  if (AppMode > 0) {return;}  // don't draw if in angle mode

  if (servoOE > 0) {
    // GREEN off
    fill(20,210,0); 
  } else {
    // RED active
    fill(255,20,0); 
  } stroke(255,255,255); strokeWeight(1); rect(682,10,11,30);
}

// ----------------------------------------------------------------------

void drawRxMsg() {
  // put text in the Rx field
  drawTxtField("                                             ", 242, 672);
  drawTxtField(msgRx, 242, 672);
}

// ----------------------------------------------------------------------

void drawServoVal() {
  // put text in Val field
  if (comFlag < 1) {return;}  // don't draw if not connected
  
  int zX = 82; String zSV$ = ""; int zSD;
  if (mDwn > 0) {DitherOffset = 0;}
  
  if (AppMode < 3) {
    zSV$ = str(servoVal[0]); if (servoTgt == 0) {if (Dither && Ch_En[0]) {
        zSD = servoVal[0] + DitherOffset;
        zSD = max(zSD,servoVLL[0]); zSD = min(zSD,servoVUL[0]);
        zSV$ = str(zSD);
      }
    }
    drawTxtField("       ", zX, 57); drawTxtField(zSV$, zX, 57);    // Ch0
  }
  
  zSV$ = str(servoVal[1]); if (servoTgt == 1) {if (Dither && Ch_En[1]) {
      zSD = servoVal[1] + DitherOffset;
      zSD = max(zSD,servoVLL[1]); zSD = min(zSD,servoVUL[1]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 95); drawTxtField(zSV$, zX, 95);    // Ch1

  zSV$ = str(servoVal[2]); if (servoTgt == 2) {if (Dither && Ch_En[2]) {
      zSD = servoVal[2] + DitherOffset;
      zSD = max(zSD,servoVLL[2]); zSD = min(zSD,servoVUL[2]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 134); drawTxtField(zSV$, zX, 134);  // Ch2

  zSV$ = str(servoVal[3]); if (servoTgt == 3) {if (Dither && Ch_En[3]) {
      zSD = servoVal[3] + DitherOffset;
      zSD = max(zSD,servoVLL[3]); zSD = min(zSD,servoVUL[3]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 172); drawTxtField(zSV$, zX, 172);  // Ch3

  zSV$ = str(servoVal[4]); if (servoTgt == 4) {if (Dither && Ch_En[4]) {
      zSD = servoVal[4] + DitherOffset;
      zSD = max(zSD,servoVLL[4]); zSD = min(zSD,servoVUL[4]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 211); drawTxtField(zSV$, zX, 211);  // Ch4

  zSV$ = str(servoVal[5]); if (servoTgt == 5) {if (Dither && Ch_En[5]) {
      zSD = servoVal[0] + DitherOffset;
      zSD = max(zSD,servoVLL[0]); zSD = min(zSD,servoVUL[0]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 249); drawTxtField(zSV$, zX, 249);  // Ch5

  zSV$ = str(servoVal[6]); if (servoTgt == 6) {if (Dither && Ch_En[6]) {
      zSD = servoVal[6] + DitherOffset;
      zSD = max(zSD,servoVLL[6]); zSD = min(zSD,servoVUL[6]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 288); drawTxtField(zSV$, zX, 288);  // Ch6

  zSV$ = str(servoVal[7]); if (servoTgt == 7) {if (Dither && Ch_En[7]) {
      zSD = servoVal[7] + DitherOffset;
      zSD = max(zSD,servoVLL[7]); zSD = min(zSD,servoVUL[7]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 326); drawTxtField(zSV$, zX, 326);  // Ch7

  zSV$ = str(servoVal[8]); if (servoTgt == 8) {if (Dither && Ch_En[8]) {
      zSD = servoVal[8] + DitherOffset;
      zSD = max(zSD,servoVLL[8]); zSD = min(zSD,servoVUL[8]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 364); drawTxtField(zSV$, zX, 364);  // Ch8

  zSV$ = str(servoVal[9]); if (servoTgt == 9) {if (Dither && Ch_En[9]) {
      zSD = servoVal[9] + DitherOffset;
      zSD = max(zSD,servoVLL[9]); zSD = min(zSD,servoVUL[9]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 402); drawTxtField(zSV$, zX, 402);  // Ch9

  zSV$ = str(servoVal[10]); if (servoTgt == 10) {if (Dither && Ch_En[10]) {
      zSD = servoVal[10] + DitherOffset;
      zSD = max(zSD,servoVLL[10]); zSD = min(zSD,servoVUL[10]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 440); drawTxtField(zSV$, zX, 440); // Ch10

  zSV$ = str(servoVal[11]); if (servoTgt == 11) {if (Dither && Ch_En[11]) {
      zSD = servoVal[11] + DitherOffset;
      zSD = max(zSD,servoVLL[11]); zSD = min(zSD,servoVUL[11]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 479); drawTxtField(zSV$, zX, 479); // Ch11

  zSV$ = str(servoVal[12]); if (servoTgt == 12) {if (Dither && Ch_En[12]) {
      zSD = servoVal[12] + DitherOffset;
      zSD = max(zSD,servoVLL[12]); zSD = min(zSD,servoVUL[12]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 518); drawTxtField(zSV$, zX, 518); // Ch12

  zSV$ = str(servoVal[13]); if (servoTgt == 13) {if (Dither && Ch_En[13]) {
      zSD = servoVal[13] + DitherOffset;
      zSD = max(zSD,servoVLL[13]); zSD = min(zSD,servoVUL[13]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 556); drawTxtField(zSV$, zX, 556); // Ch13

  zSV$ = str(servoVal[14]); if (servoTgt == 14) {if (Dither && Ch_En[14]) {
      zSD = servoVal[14] + DitherOffset;
      zSD = max(zSD,servoVLL[14]); zSD = min(zSD,servoVUL[14]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 594); drawTxtField(zSV$, zX, 594); // Ch14

  zSV$ = str(servoVal[15]); if (servoTgt == 15) {if (Dither && Ch_En[15]) {
      zSD = servoVal[15] + DitherOffset;
      zSD = max(zSD,servoVLL[15]); zSD = min(zSD,servoVUL[15]);
      zSV$ = str(zSD);
    }
  }
  drawTxtField("       ", zX, 632); drawTxtField(zSV$, zX, 632); // Ch15
}

// ----------------------------------------------------------------------

void drawTxtCtrField(String zM,int zX, int zY) {
  // called by other draw text functions. This centres the text around zX
  zX = zX - (int)(textWidth(zM)/2.0);
  stroke(BR,BG,BB); fill(BR,BG,BB); 
  rect(zX, zY, textWidth(zM), 15);
  fill(TR,TG,TB); text(zM, zX, zY+12);
}

// ----------------------------------------------------------------------

void drawTxMsg() {
  // put text in Tx field
  drawTxtField("                                       ", 498, 672);
  drawTxtField(msgTx, 498, 672);
}

// ----------------------------------------------------------------------

void drawTxtField(String zM,int zX, int zY) {
  // called by other draw text functions
  stroke(242,242,242); fill(242,242,242); 
  rect(zX, zY, 8+textWidth(zM), 14);
  fill(TR,TG,TB); text(zM, zX+4, zY+12);
}

// ----------------------------------------------------------------------

void drawULMsg() {
  // put text in UL value field
  if (comFlag < 1) {return;}  // don't draw if not connected
  int zX = 738;
  if (AppMode < 3) {drawTxtCtrField("       ", zX, 57); drawTxtCtrField(str(servoVUL[0]), zX, 57);}    // Ch0
  drawTxtCtrField("       ", zX, 95); drawTxtCtrField(str(servoVUL[1]), zX, 95);    // Ch1
  drawTxtCtrField("       ", zX, 134); drawTxtCtrField(str(servoVUL[2]), zX, 134);  // Ch2
  drawTxtCtrField("       ", zX, 172); drawTxtCtrField(str(servoVUL[3]), zX, 172);  // Ch3

  drawTxtCtrField("       ", zX, 211); drawTxtCtrField(str(servoVUL[4]), zX, 211);  // Ch4
  drawTxtCtrField("       ", zX, 249); drawTxtCtrField(str(servoVUL[5]), zX, 249);  // Ch5
  drawTxtCtrField("       ", zX, 288); drawTxtCtrField(str(servoVUL[6]), zX, 288);  // Ch6
  drawTxtCtrField("       ", zX, 326); drawTxtCtrField(str(servoVUL[7]), zX, 326);  // Ch7

  drawTxtCtrField("       ", zX, 364); drawTxtCtrField(str(servoVUL[8]), zX, 364);  // Ch8
  drawTxtCtrField("       ", zX, 402); drawTxtCtrField(str(servoVUL[9]), zX, 402);  // Ch9
  drawTxtCtrField("       ", zX, 440); drawTxtCtrField(str(servoVUL[10]), zX, 440); // Ch10
  drawTxtCtrField("       ", zX, 479); drawTxtCtrField(str(servoVUL[11]), zX, 479); // Ch11

  drawTxtCtrField("       ", zX, 518); drawTxtCtrField(str(servoVUL[12]), zX, 518); // Ch12
  drawTxtCtrField("       ", zX, 556); drawTxtCtrField(str(servoVUL[13]), zX, 556); // Ch13
  drawTxtCtrField("       ", zX, 594); drawTxtCtrField(str(servoVUL[14]), zX, 594); // Ch14
  drawTxtCtrField("       ", zX, 632); drawTxtCtrField(str(servoVUL[15]), zX, 632); // Ch15
}

// ----------------------------------------------------------------------

void exit() {
  // program is closing so reset Arduino...
  usbPortWrite("XX.");
  delay(20); // allow 20ms for data to be sent
  super.exit();
}

// ----------------------------------------------------------------------

void getCOMPort() {
  // initialise serial comms if a port exists
  if (comFlag > 0) {
    // stops an existing connection if already made
    closeCOMPort(); delay(100);
  }
  comListLength = Serial.list().length;
  if (comListLength > 0) {
    comPnt++; if (comPnt >= comListLength) {comPnt = 0;}
    // test available serial port
    comName = Serial.list()[comPnt];
    println(comName);
    try {
      usbPort = new Serial(this, comName, 115200);
      //usbPort.bufferUntil(LF);
      comFlag = 1; msgCOM = comName;
      msgTx = "USB Connected"; println(msgTx);
    } catch(Exception e) {
      comFlag = 0; msgCOM = "-Error-";
    }
  } else {
    // no serial port available
    comFlag = 0; msgCOM = "- NA -";
  } drawFlag = 1;
}

// ----------------------------------------------------------------------

void incActiveSliders() {
  // normally called from a keyboard right arrow key event.
  // It will increment any active slider value
  int zservoTgt = servoTgt;  // note current target value
  for (servoTgt = 0; servoTgt < 16; servoTgt++) {
    if (Ch_En[servoTgt]) {incXSlider();}
  }
  servoTgt = zservoTgt;  // restore target value
}

// ----------------------------------------------------------------------

void incDither() {
  // increase the dither value and send it to the Arduino
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't inc if in angle mode

  if (DitherVal < 20) {
    DitherVal++; drawFlag = 1;
    //msgTx = "SF" + str(Freq) + ".";
    //usbPortWrite(msgTx);
  }
  if (Dither) {DitherCnt = 50;  DitherPhase = 0;}  // display the change
}

// ----------------------------------------------------------------------

void incFreq() {
  // increase the PWM frequency value and send it to the Arduino
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't inc if in angle mode

  if (Freq < 100) {
    Freq++; drawFlag = 1;
    msgTx = "SF" + str(Freq) + ".";
    usbPortWrite(msgTx);
  }
}

// ----------------------------------------------------------------------

void incXSlider() {
  // increment slider value
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (servoVal[servoTgt] < servoVUL[servoTgt]) {
    servoLast = servoVal[servoTgt];
    if (keySHIFT) {servoVal[servoTgt] += 5;} else {servoVal[servoTgt]++;}
    if (servoVal[servoTgt] > servoVUL[servoTgt]) {servoVal[servoTgt] = servoVUL[servoTgt];}
    setXSlider();
    if (AppMode == 0) {
      if (Ch_En[servoTgt]) {
        msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
      } else {
        msgTx = "ST" + str(servoTgt) + "." +"SV0.";
      }
      usbPortWrite(msgTx);
    }
    if (AppMode == 1) {
      msgTx = "ST" + str(servoTgt) + "." +"SA" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 2) {
      msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 3) {
      msgTx = "LT" + str(servoTgt) + "." +"LV" + str(servoVal[servoTgt]) + ".";
      if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
    }
    valueChanged();
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void LoadDefArrays() {
  // load the calibration values into the arrays
  // if you fit a new servo you need to relax values to determine limits
  // relax value limits are LL=120, UL=590
  int zVal = 350; int zVLL = 100; int zVUL = 600;
  if (AppMode == 1) {zVal = 135; zVLL = 0; zVUL = 270;}
  if (AppMode == 2) {zVal = 1500; zVLL = 500; zVUL = 2400;}
  if (AppMode == 3) {zVal = 0; zVLL = -1000; zVUL = 1000;}
  for (int zS = 0; zS < 16; zS++) {
    // Set default values
    servoVal[zS]  = zVal; servoVLL[zS]  = zVLL; servoVUL[zS]  = zVUL;
    memVal[zS]  = zVal; memVLL[zS]  = zVLL; memVUL[zS]  = zVUL;
    memLamp[zS] = 1;
  }
}

// ----------------------------------------------------------------------

void MemHeld() {
  // user clicked on mem lamp for more than 1 second so store values
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}       // don't draw if in angle mode

  cursor(WAIT);
  memVLL[servoTgt] = servoVLL[servoTgt];
  memVal[servoTgt] = servoVal[servoTgt];
  memVUL[servoTgt] = servoVUL[servoTgt];
  memLamp[servoTgt] = 3; // set lamp blue
  mDwn = 0;              // reset the mouse down flag
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void MemPressed() {
  // user clicked on a memory lamp
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't action if in angle mode

  memState = memLamp[servoTgt];   // record the current lamp setting
  memLamp[servoTgt] = 0;          // switch OFF lamp whilst pressed
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void MemRecall() {
  // recall the channel settings from memory
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't action if in angle mode

  servoVLL[servoTgt] = memVLL[servoTgt];
  servoVal[servoTgt] = memVal[servoTgt];
  servoVUL[servoTgt] = memVUL[servoTgt];
  setSliderX();            // reposition X slider
  // reset the lamp field
  switch(memState) {
    case 1: memLamp[servoTgt] = 1; break;   // reset lamp
    case 2: memLamp[servoTgt] = 1; break;   // reset lamp
    case 3: memLamp[servoTgt] = 3; break;   // reset lamp
    case 4: memLamp[servoTgt] = 3; break;   // reset lamp
  }
  // send the recalled settings to the Arduino
  msgTx = "ST" + str(servoTgt) + ".";
  if (Ch_En[servoTgt]) {
    msgTx = msgTx +"SV" + str(servoVal[servoTgt]) + ".";
  } else {
    msgTx = msgTx +"SV0.";
    }
  usbPortWrite(msgTx); 
  msgTx = "ST" + str(servoTgt) + ".";
  msgTx = msgTx + "SL" + str(servoVLL[servoTgt]) + ".";
  usbPortWrite(msgTx); 
  msgTx = "ST" + str(servoTgt) + ".";
  msgTx = msgTx + "SU" + str(servoVUL[servoTgt]) + ".";
  usbPortWrite(msgTx); 
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void readXSlider(int zS) {
  // determine X-slider position and send value
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  servoLast = servoVal[zS];  // remember the current value
  JoyXval[zS] = mX - mXD;
  JoyXval[zS] = max(JoyXval[zS], JoyX0);
  JoyXval[zS] = min(JoyXval[zS], JoyX1);
  // normal single channel mode
  servoVal[zS] = servoVLL[zS] + (((JoyXval[zS] - JoyX0) * (servoVUL[servoTgt] - servoVLL[servoTgt])) / (JoyX1 - JoyX0));
  servoVal[zS] = max(servoVal[zS], servoVLL[zS]);
  servoVal[zS] = min(servoVal[zS], servoVUL[zS]);
  if (servoLast != servoVal[zS]) {
    servoLast = servoVal[zS];
    valueChanged();
    if (AppMode == 0) {
      // in normal 16-ch PWM mode
      if (Ch_En[zS]) {
        msgTx = "ST" + str(zS) + "." +"SV" + str(servoVal[zS]) + ".";
      } else {
        msgTx = "ST" + str(zS) + "." +"SV0.";
      }
      usbPortWrite(msgTx);
    }
    if (AppMode == 1) {
      // in 16-Ch angle mode
      msgTx = "ST" + str(zS) + "." +"SA" + str(servoVal[zS]) + ".";
      if (Ch_En[zS]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 2) {
      // in PWM mode
      msgTx = "ST" + str(zS) + "." +"SV" + str(servoVal[zS]) + ".";
      if (Ch_En[zS]) {usbPortWrite(msgTx);}
    }
    if (AppMode == 3) {
      // in LX-16A mode
      msgTx = "LT" + str(zS) + "." +"LV" + str(servoVal[zS]) + ".";
      if (Ch_En[zS]) {usbPortWrite(msgTx);}
    }
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void RstFreq() {
  // reset the PWM frequency to 50Hz or 60Hz depending on mouse click
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't action if in angle mode

if (mB == LEFT) {Freq = 50;}
  else {Freq = 60;} drawFlag = 1;
  msgTx = "SF" + str(Freq) + ".";
  usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

void sendButton(String zS) {
  // called when a button is clicked on the scrn
  if (comFlag < 1) {return;}  // don't do anything if not connected
  msgTx = zS; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendExpRq() {
  // send an EXPORT request to the Arduino
  msgTx = "SE."; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendRESET() {
  // reset defaults and send RESET msg to arduino
  servoOE = 1;      // clear OE flag as Arduino will do the same
  ClearChEn();
  ClearMemLamps();
  // the Arduino should RESET and export the data values
  msgTx = "!"; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendServoVal() {
  // send the current channels servo value
  servoLast = servoVal[servoTgt];
  if (AppMode == 0) {
    if (Ch_En[servoTgt]) {
      msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
    } else {
      msgTx = "ST" + str(servoTgt) + "." +"SV0.";
    } usbPortWrite(msgTx);
  }
  if (AppMode == 1) {
    msgTx = "ST" + str(servoTgt) + "." +"SA" + str(servoVal[servoTgt]) + ".";
    if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
  }
  if (AppMode == 2) {
    msgTx = "ST" + str(servoTgt) + "." +"SV" + str(servoVal[servoTgt]) + ".";
    if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
  }
  if (AppMode == 3) {
    msgTx = "LT" + str(servoTgt) + "." +"LV" + str(servoVal[servoTgt]) + ".";
    if (Ch_En[servoTgt]) {usbPortWrite(msgTx);}
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void SetChEn() {
  // set all channel enables the same and send data to the Arduino
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  int zS = servoTgt;
  for (servoTgt = 0; servoTgt < 16; servoTgt++) {
    Ch_En[servoTgt] = Ch_En[zS];
    sendServoVal();
  } servoTgt = zS; drawFlag = 1;
  
  countChannels();
  if ((Ch_Cnt < 1) && (Dither)) {turnDitherOFF();}
}

// ----------------------------------------------------------------------

void setLL() {
  // set lower limit from slider position
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  LastVal = servoVLL[servoTgt];
  if (mB == LEFT) {
    servoVLL[servoTgt] = servoVal[servoTgt];
    if (servoVLL[servoTgt] >= servoVUL[servoTgt]) {servoVLL[servoTgt]--;}
  } else {
    servoVLL[servoTgt] = servoLL;
  }
  setXSlider(); drawFlag = 1;
  msgTx = "SL" + str(servoVLL[servoTgt]) + ".";
  usbPortWrite(msgTx);
  if (LastVal != servoVLL[servoTgt]) {valueChanged();}
}

// ----------------------------------------------------------------------

void setLLValDwn() {
  // reduce the slider lower limit
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  if (servoVLL[servoTgt] > servoLL) {
    servoVLL[servoTgt]--;
    if (keySHIFT) {servoVLL[servoTgt] -= 5;} else {servoVLL[servoTgt]--;}
    if (servoVLL[servoTgt] < servoLL) {servoVLL[servoTgt] = servoLL;}
    setXSlider(); drawFlag = 1;
    msgTx = "SL" + str(servoVLL[servoTgt]) + ".";
    usbPortWrite(msgTx);
    valueChanged();
  }
}

// ----------------------------------------------------------------------

void setLLValUp() {
  // increase the slider upper limit
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}

  if (servoVLL[servoTgt] < servoUL) {
    if (keySHIFT) {servoVLL[servoTgt] += 5;} else {servoVLL[servoTgt]++;}
    if (servoVLL[servoTgt] > servoUL) {servoVLL[servoTgt] = servoUL;}
    if (servoVLL[servoTgt] >= servoVUL[servoTgt]) {servoVUL[servoTgt]++;}
    if (servoVal[servoTgt] < servoVLL[servoTgt]) {servoVal[servoTgt] = servoVLL[servoTgt];}
    setXSlider(); drawFlag = 1;
    msgTx = "SL" + str(servoVLL[servoTgt]) + ".";
    usbPortWrite(msgTx);
    valueChanged();
  }
}

// ----------------------------------------------------------------------

void setMode() {
  // change the mode of the application
  switch(AppMode) {
    case 0:
      // switch form to default mode
      servoLL = 50; servoUL = 600;
      LoadDefArrays(); setXSlider(); break;
    case 1:
      // switch form into angle mode
      servoLL = 0; servoUL = 270;
      LoadDefArrays(); setXSlider(); break;
    case 2:
      // switch form into PWM mode
      servoLL = 500; servoUL = 2400;
      LoadDefArrays(); setXSlider(); break;
    case 3:
      // switch form into LX-16A Serial mode
      servoLL = -1000; servoUL = 1000;
      msTimer = 0; msRun = false;  // start with timer reset
      LoadDefArrays(); setXSlider(); break;
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setSliderX() {
  // set the X-slider position for current servoVal[servoTgt] value
  if ((comFlag < 1) && (Loaded)) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  int zX0 = 259; int zX1 = 655; int zSV = servoVal[servoTgt];
  if (Dither) {
    zSV += DitherOffset;
    zSV = max(zSV,servoVLL[servoTgt]); zSV = min(zSV,servoVUL[servoTgt]);
  }
  JoyXval[servoTgt] = zX0 + ((zSV - servoVLL[servoTgt]) * (zX1 - zX0))/(servoVUL[servoTgt] - servoVLL[servoTgt]);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setUL() {
  // set upper limit from slider position if left mouse click
  // otherwise use absolute limit
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  LastVal = servoVUL[servoTgt];
  if (mB == LEFT) {
    servoVUL[servoTgt] = servoVal[servoTgt];
    if (servoVUL[servoTgt] <= servoVLL[servoTgt]) {servoVUL[servoTgt]++;}
  } else {
    servoVUL[servoTgt] = servoUL;
  }
  setXSlider(); drawFlag = 1;
  msgTx = "SU" + str(servoVUL[servoTgt]) + ".";
  usbPortWrite(msgTx);
  if (LastVal != servoVUL[servoTgt]) {valueChanged();}
}

// ----------------------------------------------------------------------

void setULValDwn() {
  // reduce the slider upper limit
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  if (servoVUL[servoTgt] > servoLL) {
    if (keySHIFT) {servoVUL[servoTgt] -= 5;} else {servoVUL[servoTgt]--;}
    if (servoVUL[servoTgt] < servoLL) {servoVLL[servoTgt] = servoLL;}
    if (servoVUL[servoTgt] <= servoVLL[servoTgt]) {servoVLL[servoTgt]--;}
    if (servoVal[servoTgt] > servoVUL[servoTgt]) {servoVal[servoTgt] = servoVUL[servoTgt];}
    setXSlider(); drawFlag = 1;
    msgTx = "SU" + str(servoVUL[servoTgt]) + ".";
    usbPortWrite(msgTx);
    valueChanged();
  }
}

// ----------------------------------------------------------------------

void setULValUp() {
  // increase the slider upper limit
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  if (servoVUL[servoTgt] < servoUL) {
    if (keySHIFT) {servoVUL[servoTgt] += 5;} else {servoVUL[servoTgt]++;}
    if (servoVUL[servoTgt] > servoUL) {servoVUL[servoTgt] = servoUL;}
    setXSlider(); drawFlag = 1;
    msgTx = "SU" + str(servoVUL[servoTgt]) + ".";
    usbPortWrite(msgTx);
    valueChanged();
  }
}

// ----------------------------------------------------------------------

void setXSlider() {
  // set the X-slider position for all servoVal[n] values
  if ((comFlag < 1) && (Loaded)) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  int zX0 = 259; int zX1 = 655;
  for (int zS = 0; zS < 16; zS++) {
    JoyXval[zS] = zX0 + ((servoVal[zS] - servoVLL[zS]) * (zX1 - zX0))/(servoVUL[zS] - servoVLL[zS]);
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleChEn() {
  // toggles the channel enable and sends data if channel is turned ON
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if ((AppMode == 3) && (servoTgt == 0)) {return;}
  
  Ch_En[servoTgt] = !Ch_En[servoTgt];
  if (AppMode > 0) {
    if (AppMode == 3) {  
      // LX-16A serisal mode
     if (!Ch_En[servoTgt]) {msgTx = "LT" + str(servoTgt) + ".LV0.";}
      else {msgTx = "LT" + str(servoTgt) + "." +"LV" + str(servoVal[servoTgt]) + ".";}
      usbPortWrite(msgTx);
    } else if (AppMode == 2) {  
      if (Ch_En[servoTgt]) {msgTx = "ST" + str(servoTgt) + ".SV" + str(servoVal[servoTgt]) + ".";}
      else {msgTx = "SD.";}  // disable the channel
      usbPortWrite(msgTx);
    } else {
      if (Ch_En[servoTgt]) {msgTx = "SO0.";} else {msgTx = "SO1.";} 
      usbPortWrite(msgTx);
    }
  } else {
    sendServoVal();
    countChannels();
    if ((Ch_Cnt < 1) && (Dither)) {turnDitherOFF();}
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleDither() {
  // called to switch Dither ON/OFF
  if (AppMode > 0) {return;} // don't action if in angle mode

  drawFlag = 1;
  if (Ch_Cnt < 1)  {turnDitherOFF(); return;}
  
  if (Dither) {turnDitherOFF();}
  else {Dither = true;}  
  Help_Message();
}

// ----------------------------------------------------------------------

void toggleMode() {
  // change the mode of the application
  AppMode++; if (AppMode > 3) {AppMode = 0;}
  setMode();
}

// ----------------------------------------------------------------------

void toggleMsTimer() {
  // user has clicked on msTimer field so start/stop timer
  if (msRun) {
    // timer is running, so stop it
    msRun = false; msTimer = millis() - msTimer;
  } else {
    // timer is not running, so start it
    msRun = true; msTimer = millis();
  }
  DFU = true;  // redraw on mouse release
}

// ----------------------------------------------------------------------

void toggleOE() {
  // toggle the OE flag and update
  if (comFlag < 1) {return;}  // don't do anything if not connected
  if (AppMode > 0) {return;}  // don't action if in angle mode

  if (servoOE > 0) {
    // turning OE ON
    servoOE = 0; sendButton("SO0.");
  }
  else {
    // turning OE OFF
    servoOE = 1; sendButton("SO1.");
    ClearChEn();  // clear channels to indicate servo pulses are OFF
    turnDitherOFF();  // remove dither if it was ON
  } drawFlag = 1;
}

// ----------------------------------------------------------------------

void turnDitherOFF() {
  // called when Dither is active to turn it OFF
  Dither = false; DitherPhase = 0; DitherOffset = 0;
  appendOffset();
}

// ----------------------------------------------------------------------

void usbPortOp() {
  // called from a timer to empty the usb buffer
  usbLen = usbBuff.length();
  if (usbLen < 1) {return;}  // nothing in buffer
  
  // we are going to send something from usbBuff
  if (usbLen <= WiFiPW) {
    // it should only take 2.8ms to send 32 bytes
    usbPort.write(usbBuff);  // send small packet
    usbBuff = ""; usbLen = 0; // send small packet and clear buffer
  } else {
    usbPort.write(usbBuff.substring(0,WiFiPW)); // send WiFiPW length packet
    usbBuff = usbBuff.substring(WiFiPW,usbBuff.length());  // shift buffer by 16 char
    usbLen = usbBuff.length();
  }
}

// ----------------------------------------------------------------------

void usbPortWrite(String zmsg) {
  // called to send text over the serial port
  // nothing is sent if comFlag or usbEn are false
  if (comFlag > 0) {
    usbBuff = usbBuff + zmsg;  // add new message to buffer
    //println(zmsg);
  } else {usbBuff = "";}  // flush the buffer if not sending
}

// ----------------------------------------------------------------------

void valueChanged() {
  // called whenever a value changes to set lamps
  // if a lamp was green it will change to yellow
  // if a lamp was blue it will change to purple
  if (comFlag < 1) {return;}  // don't do anything if not connected
  switch(memLamp[servoTgt]) {
    case 1: memLamp[servoTgt] = 2; break;
    case 3: memLamp[servoTgt] = 4; break;
  }
}
