// ######################################################################
//
//  Servo Keyboard Control File NANO
//
//  Released:  06/10/2022

//  Author: TechKnowTone
//
// ######################################################################
/*
  TERMS OF USE: This software is furnished "as is", without technical support, and
  with no warranty, expressed or implied, as to its usefulness for any purpose. In
  no event shall the author or copyright holder be liable for any claim, damages,
  or other liability, whether in an action of contract, tort or otherwise, arising
  from, out of or in connection with the software or the use or other dealings in
  the software.

  This code communicates with the Arduino NANO to control a servos via
  the USB serial interface and keyboard button presses.
  
  This version includes:
  You can now adjust slide upper and lower limits.
  Tick link active every 500ms
  
*/

// declare libraries
import processing.serial.*;
Serial usbPort;

// declare variables

int angleMode;           // 0 = degrees, 1 = microseconds
String anyStr = "";      // any temporary string 
int anyVal = 0;          // any temporary variable
int comCnt = 0;          // count down timer detects link disconnect
int comFlag = 0;         // =1 if COM serial available
int comListLength = 0;   // number of items in COM connection list
String comName = "";     // serial port COM name
int comPnt = 0;          // COM port list pointer
int cursor = 0;          // counter used to display text cursor
String data = "";
int dither0 = 0;         // dither enabled flag for servo 0
int dither1 = 0;         // dither enabled flag for servo 1
int dither2 = 0;         // dither enabled flag for servo 2
int dither3 = 0;         // dither enabled flag for servo 3
int ditherCnt = 0;       // counter used to set the speed of the dither
int ditherPhase = 0;     // phase counter for dither functions
int drawFlag = 1;        // = 1 to force screen redraw
float halfRnge = 1500.0; // calculated half range value
int Home0 = 1500;        // default home value
int Home1 = 1200;        // default home value
int Home2 = 1900;        // default home value
int Home3 = 1400;        // default home value
PImage img;              // image loaded as background
int index = 0;
int interval = 10;       // main loop interval in milliseconds
int keyBrdEnt = 0;       // > 0 if in keyboard entry mode
String keyBrdmsg = "";   // keyboard message to be sent
int keySHIFT = 0;        // > 0 if SHIFT key held
int keyVal = 0;          // keycode value
int keyDwn = 0;          // tracks keyboard up/down activity, 0 = up, 1 = down
long lastSpace;          // time when last SPACEBAR event occure
int LF = 10;             // ASCII line feed
int mDown,mON,mTO,mX, mY;// mouse flag and last x,y
boolean MEM = false;     // true if values are stored
int MEM0 = 0;            // stores servo 0 value
int MEM1 = 0;            // stores servo 1 value
int MEM2 = 0;            // stores servo 2 value
int MEM3 = 0;            // stores servo 3 value
boolean MEM_Dwn = false; // true when the MEM button is being pressed
float midPnt = 1500.0;   // calculated mid-range value
String msgCh;            // channel message sent to Arduino
String msgCOM = "None";  // COM port name
String msgHelp = "    "; // active help message
String msgRx = "";       // message received from Arduino
String msgTx = "";       // message sent to Arduino
int mXL,mYL;             // previous x,y values
int nextMillis = 0;      // loop end time in millis
int period;              // half period in draw loop units
int period2;             // quarter period in draw loop units
int phase = 0;           // current waveform phase
int pingON = 1;          // > 1 sends regular '~' pings
int repCnt = 0;          // key repeat counter
int servoMin, servoMax;  // limit constants for servo values
int servoMMD;            // limit difference constant for servo values
int servoON;             // ON/OFF flag > 0 = ON
int servoLast = 0;       // previous value sent to servo
int servoSwp;            // servo sweep flag; default = 0  OFF, 1 = ON
int servo0Val;           // servo 0 value for manual setting
int servo1Val;           // servo 1 value for manual setting
int servo2Val;           // servo 2 value for manual setting
int servo3Val;           // servo 3 value for manual setting
int servo0LL;            // servo 0 value LL for manual setting
int servo0UL;            // servo 0 value UL for manual setting
int servo1LL;            // servo 1 value LL for manual setting
int servo1UL;            // servo 1 value UL for manual setting
int servo2LL;            // servo 2 value LL for manual setting
int servo2UL;            // servo 2 value UL for manual setting
int servo3LL;            // servo 3 value LL for manual setting
int servo3UL;            // servo 3 value UL for manual setting
float sinAmp = 0.0;      // sinewave amplitude
int sine0 = 0;           // sine sweep enabled flag for servo 0
int sine1 = 0;           // sine sweep enabled flag for servo 1
int sine2 = 0;           // sine sweep enabled flag for servo 2
int sine3 = 0;           // sine sweep enabled flag for servo 3
int spaceEnt = 0;        // used to track spacebar hits
int TB = 0;              // text Blue colour component
int TG = 0;              // text Green colour component
int tick = 0;            // tick link send counter
int timeCnt = 0;         // wave count down timer
int timeCntDel = 0;      // wave count down timer delay
long timeON = 0;         // time app has been loaded in loop cycles
int TR = 0;              // text Red colour component
int upDwnFlg = 0;        // up/down state flag, 0 = down, 1 = up
int valInc = 1;          // inc/dec value used in limit adjustment
float zF;                // any temp float value

void setup() {
  // first code to run
  mON = 0; // mouse X,Y is OFF initially
  resetFlags();
  size(672,320);  // screen size must match background bitmap
  loadScrn(0); // load the default screen
  surface.setTitle("Servo_KeyBrd_Cntrl NANO");
  getCOMPort();
  // allow Arduino time to boot up
  delay(100);
  println("Ready!");
}

void draw() {
  // main loop. Comes here about 100 times a second
  if (mDown > 0) {mTO--; if (mTO < 1) {mouseRepeat(); mTO = 4;}}
  if (drawFlag > 0) {
    drawFlag = 0;
    background(img); // define background RGB colour
    drawServoON(); drawDitherON(); drawSineON();
    drawLLMsg(); drawULMsg();
    drawMEM(); drawPingLamp();
    drawCOMMsg(); drawRxMsg();
    if (keyBrdEnt < 1) {drawTxMsg();}
    else {drawKeybrdTxt(); drawFlag = 1;}
    drawHelpMsg();
  }
  if (mON > 0) {drawMouseXY();}
  if (keyBrdEnt > 0) {
    keyBrdEnt--; if (keyBrdEnt < 1) {drawFlag = 1;}
  } if (!focused) {keyBrdEnt = 0; drawFlag = 1;}
  
  // now wait for end of 10ms loop time
  while (millis() < nextMillis) {
    // wait for remainder of 10ms period to sync overall period
  } nextMillis = millis() + interval; timeON++;
  
  // perform dither shifts if any switched on
  ditherCnt++; if (ditherCnt > 20) {
    ditherCnt = 0; ditherPhase++;
    if (ditherPhase > 3) {ditherPhase = 0;}
    if (dither0 > 0) {setDither(0);}
    if (dither1 > 0) {setDither(1);}
    if (dither2 > 0) {setDither(2);}
    if (dither3 > 0) {setDither(3);}
  }
  
  // perform sinusoidal sweeps if any switched on
  if (timeCnt != period2) {
    timeCnt--;
    if (timeCnt < 1) {
      phase++; if (phase >1) {phase = 0;}
      timeCnt = period; timeCntDel = 30;
    }    
  } else {
    // peak of sinewave amplitude, so pause
    timeCntDel--; if (timeCntDel < 1) {timeCnt--;}
  }
  sinAmp = sin(((timeCnt)*3.14)/period);
  if (sine0 > 0) {setSine(0);}
  if (sine1 > 0) {setSine(1);}
  if (sine2 > 0) {setSine(2);}
  if (sine3 > 0) {setSine(3);}
  
  // send ticks to maintain USB link
  if (pingON > 0) {
    tick--; if (tick < 1) {
      tick = 20; // reset tick period to 200ms
      usbPortWrite("~"); // send a tick
    }
    if (servoON > 0) {drawFlag = 1;}
    if (comCnt > 0) {
      comCnt--;
      if (comCnt == 1) {
        // nothing received from Arduino in last 5 seconds, so close COM port
        usbPort.stop();
        comFlag = 0; msgCOM = "----"; drawFlag = 1;
      }
    }
  }
}

// ----------------------------------------------------------------------

void keyBoardIp() {
  // keyboard entry mode. To exit either press <RET> or <ESC>
  keyBrdEnt = 2000; // sustain entry window for 10 seconds
  drawFlag = 1; anyStr = str(key);
  anyStr =anyStr.toUpperCase();
  if (keyCode == ALT) {return;}
  if (keyCode == CONTROL) {return;}
  if (keyCode == SHIFT) {return;}
  switch (key) {
    case BACKSPACE:
      if (keyBrdmsg.length() > 0) {
        anyVal = keyBrdmsg.length();
        println("anyVal=" + anyVal);
        if (anyVal == 1) {
          keyBrdmsg = "";
        } else {
          keyBrdmsg = keyBrdmsg.substring(0,anyVal-1);}
      } break;
    case ENTER:
      usbPortWrite(keyBrdmsg); keyBrdmsg = ""; break;
    case RETURN:
      usbPortWrite(keyBrdmsg); keyBrdmsg = ""; break;
    case 255:
      keyBrdEnt = 0; break;
    default:
      if (keyBrdmsg.length() < 19) {keyBrdmsg = keyBrdmsg + anyStr;}
  }
  // detect incorrect key entry
  if (key == CODED) {
    switch (keyCode) {
      case  UP: keyBrdEnt = 0;
      case  RIGHT: keyBrdEnt = 0;
      case  DOWN: keyBrdEnt = 0;
      case  LEFT: keyBrdEnt = 0;
    }
  }
}

// ----------------------------------------------------------------------

void keyPressed() {
  // used to receive keyboard keys
  keyDwn = 1; 
  if (key == ESC) {key = 255;}
  keyVal = key;
  if (keyBrdEnt > 0) {keyBoardIp(); return;}
  //if (key != 32) {spaceEnt = 0;} // clear the spacebar toggle flag
  anyStr = str(key); anyStr =anyStr.toUpperCase();
  switch (anyStr) {
    case  " ": // Space Bar
      if ((timeON - lastSpace) < 100) {spaceEnt++;}
      //println(timeON - lastSpace);
      if (spaceEnt > 2) {spaceEnt = 0;}
      switch(spaceEnt) {
      case 0: msgTx = "RP0."; break;
      case 1: msgTx = "RP1."; break;
      case 2: msgTx = "RP2."; break;
      }
      usbPortWrite(msgTx); drawFlag = 1;
      lastSpace = timeON; break;
    case  "A": // A
      if (keySHIFT > 0) {msgTx = "KL.KA.";} else {msgTx = "KA.";}
      usbPortWrite(msgTx); drawFlag = 1; break;
    case  "D": // D
      if (keySHIFT > 0) {msgTx = "KL.KD.";} else {msgTx = "KD.";}
      usbPortWrite(msgTx); drawFlag = 1; break;
    case  "S": // S
      if (keySHIFT > 0) {msgTx = "KL.KS.";} else {msgTx = "KS.";}
      usbPortWrite(msgTx); drawFlag = 1; break;
    case  "W": // W
      if (keySHIFT > 0) {msgTx = "KL.KW.";} else {msgTx = "KW.";}
      usbPortWrite(msgTx); drawFlag = 1; break;
  }
  if (key == CODED) {
    switch (keyCode) {
      case  DOWN: // Down arrow
        if (keySHIFT > 0) {msgTx = "KL.Kv.";} else {msgTx = "Kv.";}
        usbPortWrite(msgTx); drawFlag = 1; break;
      case  LEFT: // Left arrow
        if (keySHIFT > 0) {msgTx = "KL.K<.";} else {msgTx = "K<.";}
        usbPortWrite(msgTx); drawFlag = 1; break;
      case  RIGHT: // Right arrow
        if (keySHIFT > 0) {msgTx = "KL.K>.";} else {msgTx = "K>.";}
        usbPortWrite(msgTx); drawFlag = 1; break;
      case  SHIFT: // SHIFT key held
        keySHIFT = 1; break;
      case  UP: // Up arrow
        if (keySHIFT > 0) {msgTx = "KL.K^.";} else {msgTx = "K^.";}
        usbPortWrite(msgTx); drawFlag = 1; break;
    }
  }
}

// ----------------------------------------------------------------------

void keyReleased() {
  // clear flags
  keyDwn = 0; keySHIFT = 0;
}

// ----------------------------------------------------------------------

void mouseDragged() {
  // mouse has been clicked and dragged on window
  mX = mouseX; mY = mouseY;
}

// ----------------------------------------------------------------------

void mouseMoved() {
  // called whenever the mouse moves
  mX = mouseX; mY = mouseY;
  returnHelpMsg();
}

// ----------------------------------------------------------------------

void mousePressed() {
  // mouse has been clicked on window
  //  mX = mouseX; mY = mouseY;
  mDown = 1; mTO = 30;
  if ((mX < 26) && (mY < 26)) {toggleMouseON();}
  if ((mX > 18) && (mX < 91) && (mY > 48) && (mY < 112)) {sendRESET();}
  
  // ensure USB connection before responding to mouse clicks
  if (comFlag > 0) {
    if ((mY > 47) && (mY < 90)) {
      // Servo PWM ON/OFF
      if ((mX > 283) && (mX < 337)) {setServoOFF();}
      if ((mX > 336) && (mX < 389)) {setServoON();}
    }
    if ((mY >= 62) && (mY <= 92)) {
      // adjust Servo 1 and Servo 2 upper limits
      if ((mX >= 134) && (mX <= 149)) {set1ULValDwn();}
      if ((mX >= 150) && (mX <= 190)) {set1ULToVal();}
      if ((mX >= 191) && (mX <= 206)) {set1ULValUp();}
      if ((mX >= 467) && (mX <= 483)) {set2ULValDwn();}
      if ((mX >= 484) && (mX <= 523)) {set2ULToVal();}
      if ((mX >= 524) && (mX <= 539)) {set2ULValUp();}
    }
    if ((mY >= 66) && (mY <= 89)) {
      // servo 1 and servo 2 ditther buttons
      if ((mX >= 208) && (mX <= 233)) {ToggleDither(1);}
      if ((mX >= 541) && (mX <= 566)) {ToggleDither(2);}
    }
    if ((mY >= 88) && (mY <= 110)) {
      // servo 1 and servo 2 sineN buttons
      if ((mX >= 208) && (mX <= 233)) {ToggleSine(1);}
      if ((mX >= 541) && (mX <= 566)) {ToggleSine(2);}
    }
    if ((mY >= 96) && (mY <= 135)) {
      // send keyboard button values
      if ((mX >= 144) && (mX <= 196)) {sendUSB("KW.");}
      if ((mX >= 477) && (mX <= 529)) {sendUSB("K^.");}
    }
    if ((mY >= 136) && (mY <= 176)) {
      // send keyboard button values
      if ((mX >= 93) && (mX <= 144)) {sendUSB("KA.");}
      if ((mX >= 145) && (mX <= 195)) {sendUSB("KS.");}
      if ((mX >= 196) && (mX <= 246)) {sendUSB("KD.");}
      if ((mX >= 427) && (mX <= 477)) {sendUSB("K<.");}
      if ((mX >= 476) && (mX <= 528)) {sendUSB("Kv.");}
      if ((mX >= 529) && (mX <= 579)) {sendUSB("K>.");}
    }
    if ((mY >= 117) && (mY <= 139)) {
      // servo 0 and servo 3 sinusoidal and ditther buttons
      if ((mX >= 271) && (mX <= 295)) {ToggleSine(3);}
      if ((mX >= 296) && (mX <= 320)) {ToggleDither(3);}
      if ((mX >= 604) && (mX <= 628)) {ToggleSine(0);}
      if ((mX >= 629) && (mX <= 652)) {ToggleDither(0);}
    }
      if ((mY >= 143) && (mY <= 173)) {
      // adjust Servo 3 upper and lower limits
      if ((mX >= 20) && (mX <= 35)) {set3ULValDwn();}
      if ((mX >= 36) && (mX <= 75)) {set3ULToVal();}
      if ((mX >= 76) && (mX <= 92)) {set3ULValUp();}
      if ((mX >= 248) && (mX <= 264)) {set3LLValDwn();}
      if ((mX >= 265) && (mX <= 304)) {set3LLToVal();}
      if ((mX >= 305) && (mX <= 320)) {set3LLValUp();}
    }
    if ((mY >= 143) && (mY <= 173)) {
      // adjust Servo 0 upper and lower limits
      if ((mX >= 352) && (mX <= 368)) {set0ULValDwn();}
      if ((mX >= 369) && (mX <= 409)) {set0ULToVal();}
      if ((mX >= 410) && (mX <= 425)) {set0ULValUp();}
      if ((mX >= 581) && (mX <= 597)) {set0LLValDwn();}
      if ((mX >= 598) && (mX <= 637)) {set0LLToVal();}
      if ((mX >= 638) && (mX <= 653)) {set0LLValUp();}
    }
    if ((mY >= 180) && (mY <= 210)) {
      // adjust Servo 1 and Servo 2 lower limits
      if ((mX >= 134) && (mX <= 149)) {set1LLValDwn();}
      if ((mX >= 150) && (mX <= 190)) {set1LLToVal();}
      if ((mX >= 191) && (mX <= 206)) {set1LLValUp();}
      if ((mX >= 467) && (mX <= 483)) {set2LLValDwn();}
      if ((mX >= 484) && (mX <= 523)) {set2LLToVal();}
      if ((mX >= 524) && (mX <= 539)) {set2LLValUp();}
    }
    if ((mY >= 194) && (mY <= 233)) {
      // send Home or Reset position demands
      if ((mX >= 285) && (mX <= 336)) {sendHomeCmd();}
      if ((mX >= 337) && (mX <= 387)) {sendRPosCmd();}
    }
  }
  if ((mX >= 564) && (mY >= 192) && (mX <= 616) && (mY <= 233)) {
    MEM_Dwn = true;
  }
  
  if ((mX >= 62) && (mX <= 150) && (mY >= 242) && (mY <= 265)) {
    if (mouseButton == LEFT) {getCOMPort();}
    else {endCOMPort();}
  }
  if ((mX >= 189) && (mX <= 472) && (mY >= 242) && (mY <= 265)) {
    GClip.copy(msgRx);
    // To paste from clipboard use: String = GClip.paste();
}
  if ((mX >= 512) && (mX <= 650) && (mY >= 242) && (mY <= 265)) {
    if (comFlag > 0) {keyBrdEnt = 2000; keyBrdmsg = ""; drawFlag = 1;}
  } else {keyBrdEnt = 0; drawFlag = 1;}
}

// ----------------------------------------------------------------------

void mouseReleased() {
  // mouse released from click and drag operation
  mX = mouseX; mY = mouseY;
  mDown = 0; repCnt = 0; valInc = 1;
  if (MEM_Dwn) {MEM_Dwn = false; MEM_Recall();}
}

// ----------------------------------------------------------------------

void mouseRepeat() {
  // called if mouse still down after a pause period
  mX = mouseX; mY = mouseY;
  // ensure USB connection before responding to mouse clicks
  if (comFlag > 0) {
    if ((mY >= 62) && (mY <= 92)) {
      // adjust Servo 1 and Servo 2 upper limits
      if ((mX >= 134) && (mX <= 149)) {set1ULValDwn();}
      if ((mX >= 191) && (mX <= 206)) {set1ULValUp();}
      if ((mX >= 467) && (mX <= 483)) {set2ULValDwn();}
      if ((mX >= 524) && (mX <= 539)) {set2ULValUp();}
    }
    if ((mY >= 96) && (mY <= 135)) {
      // send keyboard button values
      if ((mX >= 144) && (mX <= 196)) {sendUSB("KW.");}
      if ((mX >= 477) && (mX <= 529)) {sendUSB("K^.");}
    }
    if ((mY >= 136) && (mY <= 176)) {
      // send keyboard button values
      if ((mX >= 93) && (mX <= 144)) {sendUSB("KA.");}
      if ((mX >= 145) && (mX <= 195)) {sendUSB("KS.");}
      if ((mX >= 196) && (mX <= 246)) {sendUSB("KD.");}
      if ((mX >= 427) && (mX <= 477)) {sendUSB("K<.");}
      if ((mX >= 476) && (mX <= 528)) {sendUSB("Kv.");}
      if ((mX >= 529) && (mX <= 579)) {sendUSB("K>.");}
    }
    if ((mY >= 143) && (mY <= 173)) {
      // adjust Servo 3 upper and lower limits
      if ((mX >= 20) && (mX <= 35)) {set3ULValDwn();}
      if ((mX >= 76) && (mX <= 92)) {set3ULValUp();}
      if ((mX >= 248) && (mX <= 264)) {set3LLValDwn();}
      if ((mX >= 305) && (mX <= 320)) {set3LLValUp();}
    }
    if ((mY >= 143) && (mY <= 173)) {
      // adjust Servo 0 upper and lower limits
      if ((mX >= 352) && (mX <= 368)) {set0ULValDwn();}
      if ((mX >= 410) && (mX <= 425)) {set0ULValUp();}
      if ((mX >= 581) && (mX <= 597)) {set0LLValDwn();}
      if ((mX >= 638) && (mX <= 653)) {set0LLValUp();}
    }
    if ((mY >= 180) && (mY <= 210)) {
      // adjust Servo 1 and Servo 2 lower limits
      if ((mX >= 134) && (mX <= 149)) {set1LLValDwn();}
      if ((mX >= 191) && (mX <= 206)) {set1LLValUp();}
      if ((mX >= 467) && (mX <= 483)) {set2LLValDwn();}
      if ((mX >= 524) && (mX <= 539)) {set2LLValUp();}
    }
    
    if ((mX >= 564) && (mY >= 192) && (mX <= 616) && (mY <= 233)) {
      MEM_Dwn = false; MEM_Store(); mDown = 0; return;
    }

    repCnt++; if (repCnt > 10) {repCnt = 0; valInc++;}
  }
}

// ----------------------------------------------------------------------

void serialEvent(Serial usbPort) {
  // respond to a serial data received event
  int zL = 0; comCnt = 500;
  data = usbPort.readStringUntil(LF);
  anyStr = trim(data); data = "";
  if (anyStr.equals("~")) {return;}
  msgRx = anyStr; println(msgRx);
  zL = msgRx.length();
  if (zL > 5) {
    if (msgRx.substring(0,1).equals("H")) {
      if (msgRx.substring(1,3).equals("0=")) {Home0 = int(msgRx.substring(3));}
      else if (msgRx.substring(1,3).equals("1=")) {Home1 = int(msgRx.substring(3));}
      else if (msgRx.substring(1,3).equals("2=")) {Home2 = int(msgRx.substring(3));}
      else if (msgRx.substring(1,3).equals("3=")) {Home3 = int(msgRx.substring(3));}
    }
    else if (msgRx.substring(0,2).equals("SN")) {
      if (msgRx.substring(2,6).equals("0LL=")) {servo0LL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("0UL=")) {servo0UL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("1LL=")) {servo1LL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("1UL=")) {servo1UL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("2LL=")) {servo2LL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("2UL=")) {servo2UL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("3LL=")) {servo3LL = int(msgRx.substring(6));}
      else if (msgRx.substring(2,6).equals("3UL=")) {servo3UL = int(msgRx.substring(6));}
    }  
  }
  if (zL > 2) {
    // here we are receiving an SVx=yyy value from the Arduino
    // we will add to the displayed message the home offset
    if (msgRx.substring(0,2).equals("SV")) {
      if (msgRx.substring(2,3).equals("0")) {
        servo0Val = int(msgRx.substring(4));
        anyVal = servo0Val - Home0; anyStr = "H"; if (anyVal >= 0) {anyStr = "H+";}
        msgRx = msgRx + "     " + anyStr + str(anyVal);}
      else if (msgRx.substring(2,3).equals("1")) {
        servo1Val = int(msgRx.substring(4));
        anyVal = servo1Val - Home1; anyStr = "H"; if (anyVal >= 0) {anyStr = "H+";}
        msgRx = msgRx + "     " + anyStr + str(anyVal);}
      else if (msgRx.substring(2,3).equals("2")) {
        servo2Val = int(msgRx.substring(4));
        anyVal = servo2Val - Home2; anyStr = "H"; if (anyVal >= 0) {anyStr = "H+";}
        msgRx = msgRx + "     " + anyStr + str(anyVal);}
      else if (msgRx.substring(2,3).equals("3")) {
        servo3Val = int(msgRx.substring(4));
        anyVal = servo3Val - Home3; anyStr = "H"; if (anyVal >= 0) {anyStr = "H+";}
        msgRx = msgRx + "     " + anyStr + str(anyVal);}
    }
  }
  drawFlag = 1;
}    

// ----------------------------------------------------------------------
