
// ---------------------------------------------------------------------------------

void attachServos(int zDel) {
  // attach all 4 servos
  // delay used during power-up reset, otherwiswe set to 0
  servoEn = true; // set the global attached flag
  servoInst[0].attach(servoPins[0]);
  servoInst[0].writeMicroseconds(servoVal[0]);
  servoAtt[0] = true;
  delay(zDel);
  servoInst[1].attach(servoPins[1]);
  servoInst[1].writeMicroseconds(servoVal[1]);
  servoAtt[1] = true;
  delay(zDel);
  setVertMinMax(); // ensure vertical channel is within limits
  servoInst[2].attach(servoPins[2]);
  servoInst[2].writeMicroseconds(servoVal[2]);
  servoAtt[2] = true;
  delay(zDel);
  servoInst[3].attach(servoPins[3]);
  servoInst[3].writeMicroseconds(servoVal[3]);
  servoAtt[3] = true;
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // reads a key from the keyboard annd reacts accordingly
  if (zkeyVal == 10) {return;}  // ignore carriage return
  if (zkeyVal == 13) {return;}  // ignore line feed
  keyChar = char(zkeyVal);
//      Serial.print("keyVal=");
//      Serial.print(keyChar);
//      Serial.print("   ASCII=");
//      Serial.println(keyVal);
  int zNF = 0;
  switch (keyChar) {
    case '.': doCmd(); zNF = 1; break;
    case '-': cmdSgn = -1; zNF = 1; break;
    case '!': // Soft RESET
      setup(); return;    //call reset
    case '#': cmdMode = ' '; cmdType = ' '; cmdSgn = 1; cmdVal = 0; zNF = 1; break;
    case '~': cmdRec = 1; zNF = 1; break; // special ping value received
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); zNF = 1;break;
    case '1': extendCmdVal(1); zNF = 1;break;
    case '2': extendCmdVal(2); zNF = 1;break;
    case '3': extendCmdVal(3); zNF = 1;break;
    case '4': extendCmdVal(4); zNF = 1;break;
    case '5': extendCmdVal(5); zNF = 1;break;
    case '6': extendCmdVal(6); zNF = 1;break;
    case '7': extendCmdVal(7); zNF = 1;break;
    case '8': extendCmdVal(8); zNF = 1;break;
    case '9': extendCmdVal(9); zNF = 1;break;
  }
  if (zNF == 0) {
    if (cmdMode == ' ') {
      // test for new Command Mode char?
      cmdMode = keyChar;
      // test for lower case
      switch (keyChar) {
        case 'h': cmdMode = 'H'; break;
        case 'k': cmdMode = 'K'; break;
        case 'r': cmdMode = 'R'; break;
        case 's': cmdMode = 'S'; break;
      } cmdType = ' '; cmdVal = 0;
    } else {
      // test for Command Type char?
      cmdType = keyChar;
      // test for lower case
      switch (keyChar) {
//        case '<': cmdType = '<'; break;
//        case '>': cmdType = '>'; break;
//        case '^': cmdType = '^'; break;
        case 'a': cmdType = 'A'; break;
        case 'd': cmdType = 'D'; break;
        case 'f': cmdType = 'F'; break;
        case 'l': cmdType = 'L'; break;
        case 'm': cmdType = 'M'; break;
        case 's': cmdType = 'S'; break;
        case 't': cmdType = 'T'; break;
        case 'u': cmdType = 'U'; break;
        case 'v': cmdType = 'V'; break;
        case 'w': cmdType = 'W'; break;
      }
    }
  }
}

// ----------------------------------------------------------------------

void detachServos() {
  // detach all servos and clear the enabled flag
  servoEn = false;  // clear the global attached flag
  servoInst[0].detach(); servoAtt[0] = false;
  servoInst[1].detach(); servoAtt[1] = false;
  servoInst[2].detach(); servoAtt[2] = false;
  servoInst[3].detach(); servoAtt[3] = false;
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received so execute command if valid
  // Commands:
  // !       - forces a 'soft' runPOST RESET
  // HM.     - move to HOME position
  // K<.     - move arm left [IR] Servo 0
  // K>.     - move arm right [IR] Servo 0
  // K^.     - move vertical arm down, head moves up, Servo 2
  // Kv.     - move vertical arm up, head moves down, Servo 2
  // KA.     - open jaws, Servo 3
  // KD.     - close jaws, Servo 3
  // KE.     - enable all servos PWM
  // KL.     - speed limit mode, clears speed increments
  // KO.     - disable all servos PWM
  // KS.     - move arm backwards, Servo 1
  // KW.     - move arm forwards, Servo 1
  // KZ.     - dither last value to see if central
  // RM.     - move to RESET rest position
  // SAnn.   - set servo angle in degrees 0 - 180
  // SD.     - detach the active servo pin
  // SFnn.   - set the PWM frequency.
  // SLnn.   - set target servo lower limit in microseconds
  // SMnn.   - set servoMain angle in microseconds servoLL - servoUL
  // SSnn.   - set the target servo number 0 - 3, but don't feedback value
  // STnn.   - set the target servo number
  // SUnn.   - set target servo upper limit in microseconds
  // SVnn.   - set target servo angle
  // SW.     - re-writes current servo values

  int zPrint = 1; // set flag for echo Rx response
  switch (cmdMode) {
    case ' ': break;
    case 'H':
      switch (cmdType) {
        case 'M':
          Set_MainTask(2);  // goto Home position
          break;
      } break;
    case 'K':
      if (Calibrated) {setCalibrated(false);} // commands force system into uncalibrated state
      zPrint = 0;
      if (servoEn) {
        switch (cmdType) {
          case '<': if (servoEn){keyBdTurnLeft();} break;
          case '>': if (servoEn){keyBdTurnRight();} break;
          case '^': if (servoEn){keyBdVertDwn();} break;
          case 'V': if (servoEn){keyBdVertUp();} break;
          case 'A': if (servoEn){keyBdJawOpen();} break;
          case 'D': if (servoEn){keyBdJawClose();} break;
          case 'L': incS0 = 1; incS1 = 1; incS2 = 1;  incS3 = 1; break;
          case 'S': if (servoEn){keyBdArmBck();} break;
          case 'W': if (servoEn){keyBdArmFwd();} break;
        }
      }
      switch (cmdType) {
        case 'E': servoEn = true; PWR_timeout = PWR_tMax; attachServos(100); zPrint = 1; break;
        case 'O': servoEn = false; PWR_timeout = 0; detachServos(); zPrint = 1; break;
        case 'Z': keyBdDither(); break;
      } break;
    case 'R':
      switch (cmdType) {
        case 'M':
          Set_MainTask(1);  // goto Rest position
          break;
      } break;
    case 'S':
      if (Calibrated) {setCalibrated(false);} // commands force system into uncalibrated state
      switch (cmdType) {
        case ' ': break;
//        case 'A':
//          Angle = cmdVal; servoInst[servoNum].write(Angle);
//          // report the angle in microseconds
//          Serial.print("SM=");
//          Serial.print(servoInst[servoNum].readMicroseconds());
//          Serial.println("."); zPrint = 0;
//          break;
        case 'D':
          servoInst[servoNum].detach(); servoAtt[servoNum] = false;
          break;
        case 'L':
          // set the lower limit of the target servo 0 - 3
          servoMin[servoNum] = cmdVal;
          if (servoVal[servoNum] < cmdVal) {
            // if servo is outside of this limit move it
            PWR_timeout = PWR_tMax; if (!servoEn) {attachServos(0);}
            servoVal[servoNum] = cmdVal;
            if (!servoAtt[servoNum]) {servoInst[servoNum].attach(servoPins[servoNum]);}
            servoInst[servoNum].writeMicroseconds(servoVal[servoNum]);
            Serial.print(F("SV"));
            Serial.print(servoNum);
            Serial.print(F("="));
            Serial.println(servoVal[servoNum]); zPrint = 0;
          } break;
//        case 'M':
//          Angle = cmdVal; servoMain.writeMicroseconds(Angle);
//          // report the angle in degrees
//          PrintTx+= "SA=" + String(servoMain.read()) + ".\n"; zPrint = 0;
//          break;
        case 'S':
          // set the angle of the target servo 0 - 3
          // but don't report the angle to the serial port
          // ensure that it is within the current limits
          cmdVal = max(servoMin[servoNum], cmdVal);
          cmdVal = min(servoMax[servoNum], cmdVal);
          servoVal[servoNum] = cmdVal;
          PWR_timeout = PWR_tMax;
          if (Calibrated && (servoNum == 1)) {
            // it is when servo[1] is moved we auto-adjust servo[2] limits
            if (!servoAtt[2]) {servoInst[2].attach(servoPins[2]);}
            setVertMinMax(); servo2Mod = servoVal[2];
            servoInst[2].writeMicroseconds(servo2Mod);
          } 
          if (servoNum == 0) {
            incOffset = true;  // turn demand so increment thermal offset counter
            if (!servoAtt[0]) {servoInst[0].attach(servoPins[0]);}
            servoInst[0].writeMicroseconds(servoVal[0] - servoOff0);
          } else {
            if (!servoAtt[servoNum]) {servoInst[servoNum].attach(servoPins[servoNum]);}
            servoInst[servoNum].writeMicroseconds(servoVal[servoNum]);
          } break;
        case 'T':
          // set the target servo number 0 - 3
          // if the number is outside of that range it is ignored
          if ((cmdVal >= 0) && (cmdVal <= 3)) {servoNum = cmdVal;}
          break;
        case 'U':
          // set the upper limit of the target servo 0 - 3
          servoMax[servoNum] = cmdVal;
          if (servoVal[servoNum] > cmdVal) {
            // if servo is outside of this limit move it
            PWR_timeout = PWR_tMax; if (!servoEn) {attachServos(0);}
            servoVal[servoNum] = cmdVal;
            if (!servoAtt[servoNum]) {servoInst[servoNum].attach(servoPins[servoNum]);}
            servoInst[servoNum].writeMicroseconds(servoVal[servoNum]);
            Serial.print(F("SV"));
            Serial.print(servoNum);
            Serial.print(F("="));
            Serial.println(servoVal[servoNum]); zPrint = 0;
          } break;
        case 'V':
          // set the angle of the target servo 0 - 3
          // ensure that it is within the current limits
          cmdVal = max(servoMin[servoNum], cmdVal);
          cmdVal = min(servoMax[servoNum], cmdVal);
          servoVal[servoNum] = cmdVal;
          PWR_timeout = PWR_tMax; if (!servoEn) {attachServos(0);}
          if (Calibrated && (servoNum == 1)) {
            // it is when servo[1] is moved we auto-adjust servo[2]
            if (!servoAtt[2]) {servoInst[2].attach(servoPins[2]);}
            setVertMinMax(); servoInst[2].writeMicroseconds(servo2Mod);
            Serial.print(F("SV2="));
            Serial.println(servoVal[2]);
          } 
          if (servoNum == 0) {
            incOffset = true;  // turn demand so increment thermal offset counter
            if (!servoAtt[0]) {servoInst[0].attach(servoPins[0]);}
            servoInst[0].writeMicroseconds(servoVal[0] - servoOff0);
          } else {
            if (!servoAtt[servoNum]) {servoInst[servoNum].attach(servoPins[servoNum]);}
            servoInst[servoNum].writeMicroseconds(servoVal[servoNum]);
          }
          Serial.print(F("SV"));
          Serial.print(servoNum);
          Serial.println(F("="));
          Serial.println(servoVal[servoNum]); zPrint = 0;
          break;
        case 'W':
          // re-writes current servo values by setting ValChg = true
          Serial.println(F("Servo values re-written"));
          ValChg = true; break;
      } break;
  }
//  if (zPrint > 0) {
//    // is not a reporting function, echo the received command
//    PrintTx+= String(cmdMode) + String(cmdType) + String(cmdVal) + ".\n";
//  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdSgn = 1; cmdVal = 0;
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = (cmdVal * 10) + zVal;  
}

// ---------------------------------------------------------------------------------

void JoyCtrl() {
  // called from the main loop ever 20ms
  // responds to changes in joystick demands, 0 - 1023
  // a deadband of +/-32 is applied to the joystick centre positions
  // turntable S0:
  anyInt = joyRX - 512;
  if (anyInt > 32) {
    // turn to the right by reducing the servo value
    servoVal[0] -= 8; servoVal[0] = max(servoVal[0],servoMin[0]);
    ValChg = true;
  } else if (anyInt < -32) {
    // turn to the left by increasing the servo value
    servoVal[0] += 8; servoVal[0] = min(servoVal[0],servoMax[0]);
    ValChg = true;
  }
  // forward arm S1
  anyInt = joyLY - 512;
  if (anyInt > 32) {
    // pull forward arm backwards by reducing the servo value
    servoVal[1] -= 8; servoVal[1] = max(servoVal[1],servoMin[1]);
    setVertMinMax(); servo2Mod = servoVal[2]; ValChg = true;
  } else if (anyInt < -32) {
    // push the forward arm forwards by increasing the servo value
    servoVal[1] += 8; servoVal[1] = min(servoVal[1],servoMax[1]);
    setVertMinMax(); servo2Mod = servoVal[2]; ValChg = true;
  }
  // vertical arm S2
  anyInt = joyRY - 512;
  if (anyInt > 32) {
    // pull the head down by reducing the servo value to raise the vertical arm
    servoVal[2] -= 8; setVertMinMax();
    servo2Mod = servoVal[2]; ValChg = true;
  } else if (anyInt < -32) {
    // push the head up by increasing the servo value to lower the vertical arm
    servoVal[2] += 8; setVertMinMax();
    servo2Mod = servoVal[2]; ValChg = true;
  }
  // claw jaws S3
  anyInt = joyLX - 512;
  if (anyInt > 32) {
    // close the jaws by reducing the servo value
    servoVal[3] -= 16; servoVal[3] = max(servoVal[3],servoMin[3]);
    ValChg = true;
  } else if (anyInt < -32) {
    // open the jaws by increasing the servo value
    servoVal[3] += 16; servoVal[3] = min(servoVal[3],servoMax[3]);
    ValChg = true;
  }
}

// ----------------------------------------------------------------------

void keyBdArmBck() {
  // keyboard move arm backward received 
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[1] > servoMin[1]) {
    if (!servoEn) {attachServos(0);}
    servoVal[1] = servoVal[1] - incS1; incS1++; incS1 = min(incS1, 24); 
    servoVal[1] = max(servoMin[1], servoVal[1]);
    servoInst[1].writeMicroseconds(servoVal[1]);
    if (Calibrated) {
      setVertMinMax(); servoInst[2].writeMicroseconds(servo2Mod);
      Serial.print(F("SV2="));
      Serial.println(servoVal[2]);
    } 
    servoInst[1].writeMicroseconds(servoVal[1]);
    Serial.print(F("SV1="));
    Serial.println(servoVal[1]);
  }
}

// ----------------------------------------------------------------------

void keyBdArmFwd() {
  // keyboard move arm forward received 
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[1] < servoMax[1]) {
    if (!servoEn) {attachServos(0);}
    servoVal[1] = servoVal[1] + incS1; incS1++; incS1 = min(incS1, 24); 
    servoVal[1] = min(servoMax[1], servoVal[1]);
    servoInst[1].writeMicroseconds(servoVal[1]);
    if (Calibrated) {
      setVertMinMax(); servoInst[2].writeMicroseconds(servo2Mod);
      Serial.print(F("SV2="));
      Serial.println(servoVal[2]);
    } 
    servoInst[1].writeMicroseconds(servoVal[1]);
    Serial.print(F("SV1="));
    Serial.println(servoVal[1]);
  }
}

// ----------------------------------------------------------------------

void keyBdDither() {
  // dither the last value to see if it is central
  if (keyPnt >= 0) {
    // previous key demand has occured
    Serial.println(F("KZ = Dither"));
    moveInc = 5;
    for (int zP = 0; zP < 20; zP++) {
      servoInst[keyPnt].writeMicroseconds(servoVal[keyPnt] + moveInc);
      Serial.print(F("KZ = "));
      Serial.println(servoVal[keyPnt] + moveInc);
      delay(250);
      servoInst[keyPnt].writeMicroseconds(servoVal[keyPnt]);
      Serial.print(F("KZ = "));
      Serial.println(servoVal[keyPnt]);
      delay(250);
      servoInst[keyPnt].writeMicroseconds(servoVal[keyPnt] - moveInc);
      Serial.print(F("KZ = "));
      Serial.println(servoVal[keyPnt] - moveInc);
      delay(250);
      servoInst[keyPnt].writeMicroseconds(servoVal[keyPnt]);
      Serial.print(F("KZ = "));
      Serial.println(servoVal[keyPnt]);
      delay(250); moveInc++;
    }
  }
}

// ----------------------------------------------------------------------

void keyBdJawClose() {
  // keyboard close jaws received 
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[3] > servoMin[3]) {
    if (!servoEn) {attachServos(0);}
    servoVal[3] = servoVal[3] - incS3; incS3 = incS3 + 2; incS3 = min(incS3, 48); 
    servoVal[3] = max(servoMin[3], servoVal[3]);
    servoInst[3].writeMicroseconds(servoVal[3]);
    Serial.print(F("SV3="));
    Serial.println(servoVal[3]);
  }
}

// ----------------------------------------------------------------------

void keyBdJawOpen() {
  // keyboard open jaws received 
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[3] < servoMax[3]) {
    if (!servoEn) {attachServos(0);}
    servoVal[3] = servoVal[3] + incS3; incS3 = incS3 + 2; incS3 = min(incS3, 48); 
    servoVal[3] = min(servoMax[3], servoVal[3]);
    servoInst[3].writeMicroseconds(servoVal[3]);
    Serial.print(F("SV3="));
    Serial.println(servoVal[3]);
  }
}

// ----------------------------------------------------------------------

void keyBdTurnLeft() {
  // keyboard left received
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[0] < servoMax[0]) {
    if (!servoEn) {attachServos(0);}
    incOffset = 10; // turn demand so increment thermal offset counter
    servoVal[0] = servoVal[0] + incS0; incS0++; incS0 = min(incS0, 24); 
    servoVal[0] = min(servoMax[0], servoVal[0]);
    servoInst[0].writeMicroseconds(servoVal[0] + servoOff0);
    Serial.print(F("SV0="));
    Serial.println(servoVal[0]);
  }
}

// ----------------------------------------------------------------------

void keyBdTurnRight() {
  // keyboard right received  
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[0] > servoMin[0]) {
    if (!servoEn) {attachServos(0);}
    incOffset = 10; // turn demand so increment thermal offset counter
    servoVal[0] = servoVal[0] - incS0; incS0++; incS0 = min(incS0, 24); 
    servoVal[0] = max(servoMin[0], servoVal[0]);
    servoInst[0].writeMicroseconds(servoVal[0] + servoOff0);
    Serial.print(F("SV0="));
    Serial.println(servoVal[0]);
  }
}

// ----------------------------------------------------------------------

void keyBdVertDwn() {
  // keyboard move head up
  if (Calibrated) {setVertMinMax();} // ensure limits are correct with respect to arm position
  else {servoMax[2] = servoMaxPWM;}  // not calibrated so use high limit
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[2] < servoMax[2]) {
    if (!servoEn) {attachServos(0);}
    servoVal[2] = servoVal[2] + incS2; incS2++; incS2 = min(incS2, 24); 
    servoVal[2] = min(servoMax[2], servoVal[2]);
    servoInst[2].writeMicroseconds(servoVal[2]);
    Serial.print(F("SV2="));
    Serial.println(servoVal[2]);
  }
}

// ----------------------------------------------------------------------

void keyBdVertUp() {
  // keyboard move head down
  if (Calibrated) {setVertMinMax();} // ensure limits are correct with respect to arm position
  else {servoMin[2] = servoMinPWM;}  // not calibrated so use low limit
  PWR_timeout = 3000; // preset timer for auto servo power removal after 30 sec
  if (servoVal[2] > servoMin[2]) {
    if (!servoEn) {attachServos(0);}
    servoVal[2] = servoVal[2] - incS2; incS2++; incS2 = min(incS2, 24); 
    servoVal[2] = max(servoMin[2], servoVal[2]);
    servoInst[2].writeMicroseconds(servoVal[2]);
    Serial.print(F("SV2="));
    Serial.println(servoVal[2]);
  }
}

// ---------------------------------------------------------------------------------

void LED_Tasks() {
  // performa a range of LED tasks
  // called every 20ms from the main loop
  if (LED_Del > 0) {LED_Del--; return;} // skip cycles if LED_Del > 0
  
  if (servoEn) {if (LED_Mode != 1) {set_LED_Mode(1);}}
  else if (LED_Mode != 0) {set_LED_Mode(0);}
  switch (LED_Mode) {
    case 0: LED_Blue_Mood(); break;
    case 1: LED_Warning(); break;
  }
}

// ---------------------------------------------------------------------------------

void LED_Blue_Mood() {
  // default LED mode, displays a soothing blue coloured pulse  
  FastLED.clear();
  if (LED_Inc > 0) {
    LED_Cnt++; if (LED_Cnt >= 16) {LED_Inc = -1;}
  } else {
    LED_Cnt--; if (LED_Cnt < 1) {LED_Inc = 1;}
  }
  LED[0].setRGB(0,0,LED_Cnt); LED[1].setRGB(0,0,LED_Cnt);
  LED_Del = 5; // pause for 200ms
  FastLED.show();
}

// ---------------------------------------------------------------------------------

void LED_Warning() {
  // this function toggles the left and right-hand LEDs on/off in red
  FastLED.clear();
  switch (LED_Cnt) {
    case 0: LED[0].setRGB(32,0,0); break;
    case 1: LED[1].setRGB(32,0,0); break;
  }
  LED_Cnt++; if (LED_Cnt > 1) {LED_Cnt = 0;}  // cycle round
  LED_Del = 10; // pause for 200ms
  FastLED.show();
}

// --------------------------------------------------------------------------------

void readJoyLX() {
  // reads the left X joystick and leaves the mux pointing at left Y
  // due to the joystick orientation in the case, their values are effectively reversed
  anyInt = joyLX; // note the previous value
  joyLX = analogRead(JoyLXpin);
//  if (abs(anyInt - joyLX) > 1) {Serial.print(F("JoyLX=")); Serial.println(joyLX);}
  anyInt = analogRead(JoyLYpin);
}

// --------------------------------------------------------------------------------

void readJoyLY() {
  // reads the left Y joystick and leaves the mux pointing at right X
  // due to the joystick orientation in the case, their values are effectively reversed
  anyInt = joyLY; // note the previous value
  joyLY = analogRead(JoyLYpin);
//  if (abs(anyInt - joyLY) > 1) {Serial.print(F("JoyLY=")); Serial.println(joyLY);}
  anyInt = analogRead(JoyRXpin);
}

// --------------------------------------------------------------------------------

void readJoyRX() {
  // reads the right X joystick and leaves the mux pointing at right Y
  // due to the joystick orientation in the case, their values are effectively reversed
  anyInt = joyRX; // note the previous value
  joyRX = 1023 - analogRead(JoyRXpin);
//  if (abs(anyInt - joyRX) > 1) {Serial.print(F("JoyRX=")); Serial.println(joyRX);}
  anyInt = analogRead(JoyRYpin);
}

// --------------------------------------------------------------------------------

void readJoyRY() {
  // reads the right Y joystick and leaves the mux pointing at left X
  // due to the joystick orientation in the case, their values are effectively reversed
  anyInt = joyRY; // note the previous value
  joyRY = 
  analogRead(JoyRYpin);
//  if (abs(anyInt - joyRY) > 1) {Serial.print(F("JoyRY=")); Serial.println(joyRY);}
  anyInt = analogRead(JoyLXpin);
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {decodeKey(keyVal);}
}

// ---------------------------------------------------------------------------------

void readSW0() {
  // read button switch SW0
  // called from the main loop every 10ms
  sw0State = digitalRead(sw0Pin); // record button state
//  if (sw0_Nop) {return;}          // block this function whilst == true
  
//  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
      Serial.println("SW0 Hi-LO");
      sw0Cnt++;       // count on the falling edge
      sw0DwnTime = 0; // restart button down time counter
      sw0Timer = 0;   // reset the timer
      Set_MainTask(3);  // goto Floor position
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      sw0DwnTime++; // track button pressed time
      if (sw0Timer >= 100) {
        // button held down for >= 1 seconds, a long press
        sw0Cnt = 0; sw0Timer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0LastState == LOW) {
      // SW0 has just been released
      Serial.println("SW0 Lo-Hi");
    }
    if (sw0Timer >= 100) {
      // button released for 1 sec so assume valid button counts
      Serial.print(F("sw0Cnt = ")); Serial.println(sw0Cnt);
      sw0Cnt = 0; sw0Timer = 0; SyncTimers();
    }
  }
  sw0LastState = sw0State; // record current state
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW1() {
  // read button switch SW1
  // called from the main loop every 10ms
  sw1State = digitalRead(sw1Pin); // record button state
//  if (sw1_Nop) {return;}          // block this function whilst == true
  
//  bool sw1Clr = false;
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
      Serial.println("SW1 Hi-LO");
      sw1Cnt++;         // count on the falling edge
      sw1DwnTime = 0;   // restart button down time counter
      sw1Timer = 0;     // reset the timer
      Set_MainTask(2);  // goto Home position
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      sw1DwnTime++; // track button pressed time
      if (sw1Timer >= 100) {
        // button held down for >= 1 seconds, a long press
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // SW1 button is released
    //
    //##############################################################################
    if (sw1LastState == LOW) {
      // SW0 has just been released
      Serial.println("SW1 Lo-Hi");
    }
    if (sw1Timer >= 100) {
      // button released for 1 sec so assume valid button counts
      Serial.print(F("sw1Cnt = ")); Serial.println(sw1Cnt);
      sw1Cnt = 0; sw1Timer = 0; SyncTimers();
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW2() {
  // read button switch SW2
  // called from the main loop every 10ms
  sw2State = digitalRead(sw2Pin); // record button state
//  if (sw2_Nop) {return;}          // block this function whilst == true
  
//  bool sw2Clr = false;
  if (sw2State == LOW) {
    //##############################################################################
    //
    // SW2 button is pressed down
    //
    //##############################################################################
    if (sw2LastState == HIGH) {
      // SW1 has just been pressed down
      Serial.println("SW2 Hi-LO");
      sw2Cnt++;         // count on the falling edge
      sw2DwnTime = 0;   // restart button down time counter
      sw2Timer = 0;     // reset the timer
      Set_MainTask(1);  // goto Rest position
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      sw2DwnTime++; // track button pressed time
      if (sw2Timer >= 100) {
        // button held down for >= 1 seconds, a long press
        sw2Cnt = 0; sw2Timer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // SW2 button is released
    //
    //##############################################################################
    if (sw2LastState == LOW) {
      // SW2 has just been released
      Serial.println("SW2 Lo-Hi");
    }
    if (sw2Timer >= 100) {
      // button released for 1 sec so assume valid button counts
      Serial.print(F("sw2Cnt = ")); Serial.println(sw2Cnt);
      sw2Cnt = 0; sw2Timer = 0; SyncTimers();
    }
  }
  sw2LastState = sw2State; // record current state
  if (sw2Cnt > 0) {sw2Timer++;}
}

// ---------------------------------------------------------------------------------

void readSwJL() {
  // read left joystick switch
  // called from the main loop every 10ms
  swJLState = digitalRead(swJoyL); // record button state
//  if (swJL_Nop) {return;}          // block this function whilst == true
  
//  bool swJLClr = false;
  if (swJLState == LOW) {
    //##############################################################################
    //
    // left joystick is pressed down
    //
    //##############################################################################
    if (swJLLastState == HIGH) {
      // switch has just been pressed down
      Serial.println("SwJL Hi-LO");
      swJLCnt++;       // count on the falling edge
      swJLDwnTime = 0; // restart button down time counter
      swJLTimer = 0;   // reset the timer
      Set_MainTask(4);  // goto Home left position
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      swJLDwnTime++; // track button pressed time
      if (swJLTimer >= 100) {
        // button held down for >= 1 seconds, a long press
        swJLCnt = 0; swJLTimer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // left joystick button is released
    //
    //##############################################################################
    if (swJLLastState == LOW) {
      // switch has just been released
      Serial.println("SwJL Lo-Hi");
    }
    if (swJLTimer >= 100) {
      // button released for 1 sec so assume valid button counts
      Serial.print(F("swJLCnt = ")); Serial.println(swJLCnt);
      swJLCnt = 0; swJLTimer = 0; SyncTimers();
    }
  }
  swJLLastState = swJLState; // record current state
  if (swJLCnt > 0) {swJLTimer++;}
}

// ---------------------------------------------------------------------------------

void readSwJR() {
  // read right joystick switch
  // called from the main loop every 10ms
  swJRState = digitalRead(swJoyR);  // record button state
//  if (swJR_Nop) {return;}          // block this function whilst == true
  
//  bool swJRClr = false;
  if (swJRState == LOW) {
    //##############################################################################
    //
    // right joystick is pressed down
    //
    //##############################################################################
    if (swJRLastState == HIGH) {
      // switch has just been pressed down
      Serial.println("SwJR Hi-LO");
      swJRCnt++;       // count on the falling edge
      swJRDwnTime = 0; // restart button down time counter
      swJRTimer = 0;   // reset the timer
      Set_MainTask(5);  // goto Home right position
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      swJRDwnTime++; // track button pressed time
      if (swJRTimer >= 100) {
        // button held down for >= 1 seconds, a long press
        swJRCnt = 0; swJRTimer = 0;
      }
    }
  } else {
    //##############################################################################
    //
    // right joystick button is released
    //
    //##############################################################################
    if (swJRLastState == LOW) {
      // switch has just been released
      Serial.println("SwJR Lo-Hi");
    }
    if (swJRTimer >= 100) {
      // button released for 1 sec so assume valid button counts
      Serial.print(F("swJRCnt = ")); Serial.println(swJRCnt);
      swJRCnt = 0; swJRTimer = 0; SyncTimers();
    }
  }
  swJRLastState = swJRState; // record current state
  if (swJRCnt > 0) {swJRTimer++;}
}

// ---------------------------------------------------------------------------------

void setCalibrated(bool zS) {
  // sets the Calibrated flag and assocxiated limits
  // if Calibrated == true then auto-limts of servo 2 will apply
  Calibrated = zS;
  if (Calibrated) {
    servoCtr[0] = Reset0; servoCtr[1] = Reset1;
    servoCtr[2] = Reset2; servoCtr[3] = Reset3;
    servoVal[0] = Reset0; servoVal[1] = Reset1;
    servoVal[2] = Reset2; servoVal[3] = Reset3;
  } else {
    Serial.println(F("Not calibrated!"));
    Serial.println(F("Setting max limits..."));
    for (int zP = 0;zP < 4;zP++) {
      servoMax[zP] = 2400; servoMin[zP] = 500; servoCtr[zP] = 1500;
      servoVal[zP] = 1500;
    }
  }
}

// ---------------------------------------------------------------------------------

void set_LED_Mode(int zM) {
  // called to change the current LED_Mode value
  if (zM != LED_Mode) {LED_Del = 0; LED_Cnt = 0; LED_Inc = 1;}
  LED_Mode = zM;
}

// ----------------------------------------------------------------------

void Set_MainTask(int zM) {
  // called to initialise a main task
  SubDel = 0;   // SubTask delay counter
  SubTask = 0;  // MainTask subtask pointer
  
  MainTask = zM;  
}

// ----------------------------------------------------------------------

void setVertMinMax() {
  // when in Calibrated state this function
  // sets the min/max limits for the vertical arm based on the 
  // angular position of the forward arm
  if (!Calibrated) {return;}
  
  long zVal;
  if (servoVal[1] > fwdArmVert) {
    // determine max limit in forward leaning posture
    zVal = servoVal[1] - fwdArmVert;
    zVal = zVal * (vertArmMaxB - vertArmMaxA);
    zVal = zVal/(servoMax[1] - fwdArmVert);
    servoMax[2] = vertArmMaxB - zVal;
    // determine min limit
    zVal = servoVal[1] - fwdArmVert;
    zVal = zVal * (vertArmMinB - vertArmMinA);
    zVal = zVal/(servoMax[1] - fwdArmVert);
    servoMin[2] = vertArmMinB - zVal;
  } else {
    servoMax[2] = vertArmMaxB; // this is abs max figure for vertical arm
    // determine min limit
    zVal = servoVal[1] - servoMin[1];
    zVal = zVal * (vertArmMinC - vertArmMinB);
    zVal = zVal/(fwdArmVert - servoMin[1]);
    servoMin[2] = vertArmMinC - zVal;
  }
  // force value to be within limits
  servoVal[2] = max(servoVal[2],servoMin[2]);
  servoVal[2] = min(servoVal[2],servoMax[2]);
}

// ---------------------------------------------------------------------------------
